from __future__ import annotations

from typing import Dict, Generator, Tuple

import netsquid as ns
from netqasm.sdk.build_epr import SER_RESPONSE_KEEP_IDX_GOODNESS, SER_RESPONSE_KEEP_LEN
from netsquid.nodes import Node
from netsquid.protocols import Protocol
from netsquid.qubits import ketstates
from netsquid_magic.link_layer import (
    MagicLinkLayerProtocolWithSignaling,
    SingleClickTranslationUnit,
)
from netsquid_magic.magic_distributor import PerfectStateMagicDistributor

from pydynaa import EventExpression
from qoala.lang.iqoala import IqoalaProgram, ProgramMeta
from qoala.runtime.config import GenericQDeviceConfig
from qoala.runtime.environment import (
    GlobalEnvironment,
    GlobalNodeInfo,
    LocalEnvironment,
)
from qoala.runtime.program import ProgramInput, ProgramInstance, ProgramResult
from qoala.runtime.schedule import ProgramTaskList
from qoala.sim.build import build_generic_qprocessor
from qoala.sim.egp import EgpProtocol
from qoala.sim.egpmgr import EgpManager
from qoala.sim.memmgr import MemoryManager
from qoala.sim.memory import ProgramMemory, SharedMemory, Topology, UnitModule
from qoala.sim.message import Message
from qoala.sim.netstackcomp import NetstackComponent
from qoala.sim.netstackinterface import NetstackInterface
from qoala.sim.netstackprocessor import NetstackProcessor
from qoala.sim.process import IqoalaProcess
from qoala.sim.qdevice import GenericPhysicalQuantumMemory, QDevice, QDeviceType
from qoala.sim.requests import (
    EprCreateType,
    NetstackCreateRequest,
    NetstackReceiveRequest,
)
from qoala.util.tests import has_multi_state


def perfect_generic_qdevice(node_name: str, num_qubits: int) -> QDevice:
    cfg = GenericQDeviceConfig.perfect_config(num_qubits=num_qubits)
    processor = build_generic_qprocessor(name="processor", cfg=cfg)
    node = Node(name=node_name, qmemory=processor)
    return QDevice(
        node=node,
        typ=QDeviceType.GENERIC,
        memory=GenericPhysicalQuantumMemory(num_qubits),
    )


def create_process(pid: int, unit_module: UnitModule) -> IqoalaProcess:
    program = IqoalaProgram(
        instructions=[], subroutines={}, meta=ProgramMeta.empty("prog")
    )
    instance = ProgramInstance(
        pid=pid,
        program=program,
        inputs=ProgramInput({}),
        tasks=ProgramTaskList.empty(program),
    )
    mem = ProgramMemory(pid=pid, unit_module=unit_module)

    process = IqoalaProcess(
        prog_instance=instance,
        prog_memory=mem,
        csockets={},
        epr_sockets=program.meta.epr_sockets,
        subroutines=program.subroutines,
        requests={},
        result=ProgramResult(values={}),
    )
    return process


def setup_components_generic(
    num_qubits: int,
) -> Tuple[NetstackProcessor, NetstackProcessor]:
    # TODO: SUPPORT ANY TOPOLOGY (ALSO REQUIRES REFACTORING CONFIG HANDLING)
    alice_qdevice = perfect_generic_qdevice("alice", num_qubits)
    bob_qdevice = perfect_generic_qdevice("bob", num_qubits)

    alice_node = alice_qdevice._node
    bob_node = bob_qdevice._node

    env = GlobalEnvironment()
    env.add_node(alice_node.ID, GlobalNodeInfo(alice_node.name, alice_node.ID))
    env.add_node(bob_node.ID, GlobalNodeInfo(bob_node.name, bob_node.ID))

    alice_comp = NetstackComponent(node=alice_node, global_env=env)
    bob_comp = NetstackComponent(node=bob_node, global_env=env)

    alice_comp.peer_out_port("bob").connect(bob_comp.peer_in_port("alice"))
    alice_comp.peer_in_port("bob").connect(bob_comp.peer_out_port("alice"))

    alice_memmgr = MemoryManager(alice_node.name, alice_qdevice)
    bob_memmgr = MemoryManager(bob_node.name, bob_qdevice)

    alice_egpmgr = EgpManager()
    bob_egpmgr = EgpManager()

    alice_intf = NetstackInterface(
        alice_comp,
        LocalEnvironment(env, alice_node.ID),
        alice_qdevice,
        alice_memmgr,
        alice_egpmgr,
    )
    bob_intf = NetstackInterface(
        bob_comp,
        LocalEnvironment(env, bob_node.ID),
        bob_qdevice,
        bob_memmgr,
        bob_egpmgr,
    )

    alice_processor = NetstackProcessor(alice_intf)
    bob_processor = NetstackProcessor(bob_intf)

    return (alice_processor, bob_processor)


def create_egp_protocols(node1: Node, node2: Node) -> Tuple[EgpProtocol, EgpProtocol]:
    link_dist = PerfectStateMagicDistributor(nodes=[node1, node2], state_delay=1000.0)
    link_prot = MagicLinkLayerProtocolWithSignaling(
        nodes=[node1, node2],
        magic_distributor=link_dist,
        translation_unit=SingleClickTranslationUnit(),
    )
    return EgpProtocol(node1, link_prot), EgpProtocol(node2, link_prot)


def create_netstack_create_request(remote_id: int) -> NetstackCreateRequest:
    return NetstackCreateRequest(
        remote_id=remote_id,
        epr_socket_id=0,
        typ=EprCreateType.CREATE_KEEP,
        num_pairs=1,
        fidelity=0.75,
        virt_qubit_ids=[0],
        result_array_addr=0,
    )


def create_netstack_receive_request(remote_id: int) -> NetstackReceiveRequest:
    return NetstackReceiveRequest(
        remote_id=remote_id,
        epr_socket_id=0,
        typ=EprCreateType.CREATE_KEEP,
        num_pairs=1,
        fidelity=0.75,
        virt_qubit_ids=[0],
        result_array_addr=0,
    )


def test_1():
    alice_processor, bob_processor = setup_components_generic(num_qubits=2)

    topology = Topology(comm_ids={0}, mem_ids={1})
    unit_module = UnitModule.from_topology(topology)

    alice_node = alice_processor._interface._comp.node
    bob_node = bob_processor._interface._comp.node

    num_pairs = 3
    fidelity = 0.75
    result_array_addr = 0

    alice_request = NetstackCreateRequest(
        remote_id=bob_node.ID,
        epr_socket_id=0,
        typ=EprCreateType.CREATE_KEEP,
        num_pairs=num_pairs,
        fidelity=fidelity,
        virt_qubit_ids=[1],
        result_array_addr=result_array_addr,
    )
    bob_request = NetstackReceiveRequest(
        remote_id=alice_node.ID,
        epr_socket_id=0,
        typ=EprCreateType.CREATE_KEEP,
        num_pairs=num_pairs,
        fidelity=fidelity,
        virt_qubit_ids=[1],
        result_array_addr=result_array_addr,
    )

    alice_egpmgr = alice_processor._interface.egpmgr
    bob_egpmgr = bob_processor._interface.egpmgr

    alice_egp, bob_egp = create_egp_protocols(alice_node, bob_node)
    alice_egpmgr.add_egp(bob_node.ID, alice_egp)
    bob_egpmgr.add_egp(alice_node.ID, bob_egp)

    class NetstackProcessorProtocol(Protocol):
        def __init__(self, name: str, processor: NetstackProcessor) -> None:
            super().__init__(name)
            self._processor = processor
            self._memmgr = processor._interface.memmgr
            self._pid = 0
            self._process = create_process(self._pid, unit_module)
            self._memmgr.add_process(self._process)

        @property
        def pid(self) -> int:
            return self._pid

        @property
        def memmgr(self) -> MemoryManager:
            return self._memmgr

    class AliceProtocol(NetstackProcessorProtocol):
        def run(self) -> Generator[EventExpression, None, None]:
            yield from self._processor._interface.receive_peer_msg("bob")
            yield from self._processor.create_single_pair(
                self._process, alice_request, virt_id=0
            )

    class BobProtocol(NetstackProcessorProtocol):
        def run(self) -> Generator[EventExpression, None, None]:
            self._processor._interface.send_peer_msg("alice", Message("ready"))
            yield from self._processor.receive_single_pair(
                self._process, bob_request, virt_id=0
            )

    alice = AliceProtocol("alice", alice_processor)
    alice_processor._interface.start()  # also starts peer listeners
    alice.start()
    alice_egp.start()

    bob = BobProtocol("bob", bob_processor)
    bob_processor._interface.start()  # also starts peer listeners
    bob.start()
    bob_egp.start()

    link_prot = alice_egp._ll_prot  # same as bob_egp._ll_prot
    link_prot.start()

    ns.sim_run()

    assert alice.memmgr.phys_id_for(alice.pid, 0) == 0
    assert bob.memmgr.phys_id_for(bob.pid, 0) == 0


def test_2():
    alice_processor, bob_processor = setup_components_generic(num_qubits=2)

    topology = Topology(comm_ids={0}, mem_ids={1})
    unit_module = UnitModule.from_topology(topology)

    alice_node = alice_processor._interface._comp.node
    bob_node = bob_processor._interface._comp.node

    num_pairs = 1
    fidelity = 0.75
    result_array_addr = 0

    alice_request = NetstackCreateRequest(
        remote_id=bob_node.ID,
        epr_socket_id=0,
        typ=EprCreateType.CREATE_KEEP,
        num_pairs=num_pairs,
        fidelity=fidelity,
        virt_qubit_ids=[0],
        result_array_addr=result_array_addr,
    )
    bob_request = NetstackReceiveRequest(
        remote_id=alice_node.ID,
        epr_socket_id=0,
        typ=EprCreateType.CREATE_KEEP,
        num_pairs=num_pairs,
        fidelity=fidelity,
        virt_qubit_ids=[0],
        result_array_addr=result_array_addr,
    )

    alice_egpmgr = alice_processor._interface.egpmgr
    bob_egpmgr = bob_processor._interface.egpmgr

    alice_egp, bob_egp = create_egp_protocols(alice_node, bob_node)
    alice_egpmgr.add_egp(bob_node.ID, alice_egp)
    bob_egpmgr.add_egp(alice_node.ID, bob_egp)

    class NetstackProcessorProtocol(Protocol):
        def __init__(self, name: str, processor: NetstackProcessor) -> None:
            super().__init__(name)
            self._processor = processor
            self._memmgr = processor._interface.memmgr
            self._pid = 0
            self._process = create_process(self._pid, unit_module)
            self._memmgr.add_process(self._process)
            self._process.prog_memory.shared_mem.init_new_array(result_array_addr, 10)

        @property
        def pid(self) -> int:
            return self._pid

        @property
        def process(self) -> IqoalaProcess:
            return self._process

        @property
        def memmgr(self) -> MemoryManager:
            return self._memmgr

        @property
        def shared_mem(self) -> SharedMemory:
            return self.process.prog_memory.shared_mem

    class AliceProtocol(NetstackProcessorProtocol):
        def run(self) -> Generator[EventExpression, None, None]:
            yield from self._processor._interface.receive_peer_msg("bob")
            yield from self._processor.handle_create_ck_request(
                self._process, alice_request
            )

    class BobProtocol(NetstackProcessorProtocol):
        def run(self) -> Generator[EventExpression, None, None]:
            self._processor._interface.send_peer_msg("alice", Message("ready"))
            yield from self._processor.handle_receive_ck_request(
                self._process, bob_request
            )

    alice = AliceProtocol("alice", alice_processor)
    alice_processor._interface.start()  # also starts peer listeners
    alice.start()
    alice_egp.start()

    bob = BobProtocol("bob", bob_processor)
    bob_processor._interface.start()  # also starts peer listeners
    bob.start()
    bob_egp.start()

    link_prot = alice_egp._ll_prot  # same as bob_egp._ll_prot
    link_prot.start()

    assert (
        alice.shared_mem.get_array_part(
            result_array_addr,
            SER_RESPONSE_KEEP_LEN * 0 + SER_RESPONSE_KEEP_IDX_GOODNESS,
        )
        is None
    )

    ns.sim_run()

    assert alice.memmgr.phys_id_for(alice.pid, 0) == 0
    assert bob.memmgr.phys_id_for(bob.pid, 0) == 0

    assert (
        alice.shared_mem.get_array_part(
            result_array_addr,
            SER_RESPONSE_KEEP_LEN * 0 + SER_RESPONSE_KEEP_IDX_GOODNESS,
        )
        is not None
    )

    alice_qubit = alice_processor.qdevice.get_local_qubit(0)
    bob_qubit = bob_processor.qdevice.get_local_qubit(0)
    assert has_multi_state([alice_qubit, bob_qubit], ketstates.b00)


def test_3():
    alice_processor, bob_processor = setup_components_generic(num_qubits=3)

    topology = Topology(comm_ids={0, 1, 2}, mem_ids={0, 1, 2})
    unit_module = UnitModule.from_topology(topology)

    alice_node = alice_processor._interface._comp.node
    bob_node = bob_processor._interface._comp.node

    result_array_addr = 0

    alice_requests = {
        0: create_netstack_create_request(remote_id=bob_node.ID),
        1: create_netstack_create_request(remote_id=bob_node.ID),
    }
    bob_requests = {
        0: create_netstack_receive_request(remote_id=alice_node.ID),
        1: create_netstack_receive_request(remote_id=alice_node.ID),
    }

    alice_memmgr = alice_processor._interface.memmgr
    bob_memmgr = bob_processor._interface.memmgr

    alice_egpmgr = alice_processor._interface.egpmgr
    bob_egpmgr = bob_processor._interface.egpmgr

    alice_egp, bob_egp = create_egp_protocols(alice_node, bob_node)
    alice_egpmgr.add_egp(bob_node.ID, alice_egp)
    bob_egpmgr.add_egp(alice_node.ID, bob_egp)

    class NetstackProcessorProtocol(Protocol):
        def __init__(
            self,
            name: str,
            processor: NetstackProcessor,
            processes: Dict[int, IqoalaProcess],
        ) -> None:
            super().__init__(name)
            self._processor = processor
            self._memmgr = processor._interface.memmgr
            self._processes = processes

        @property
        def processes(self) -> Dict[int, IqoalaProcess]:
            return self._processes

        @property
        def memmgr(self) -> MemoryManager:
            return self._memmgr

    class AliceProtocol(NetstackProcessorProtocol):
        def run(self) -> Generator[EventExpression, None, None]:
            for pid, process in self.processes.items():
                yield from self._processor._interface.receive_peer_msg("bob")
                yield from self._processor.handle_create_ck_request(
                    process, alice_requests[pid]
                )

    class BobProtocol(NetstackProcessorProtocol):
        def run(self) -> Generator[EventExpression, None, None]:
            for pid, process in self.processes.items():
                self._processor._interface.send_peer_msg("alice", Message("ready"))
                yield from self._processor.handle_receive_ck_request(
                    process, bob_requests[pid]
                )

    alice_process0 = create_process(pid=0, unit_module=unit_module)
    alice_process1 = create_process(pid=1, unit_module=unit_module)
    alice_process0.shared_mem.init_new_array(0, 10)
    alice_process1.shared_mem.init_new_array(0, 10)
    alice_memmgr.add_process(alice_process0)
    alice_memmgr.add_process(alice_process1)
    alice = AliceProtocol(
        "alice", alice_processor, {0: alice_process0, 1: alice_process1}
    )
    alice_processor._interface.start()  # also starts peer listeners
    alice.start()
    alice_egp.start()

    bob_process0 = create_process(pid=0, unit_module=unit_module)
    bob_process1 = create_process(pid=1, unit_module=unit_module)
    bob_process0.shared_mem.init_new_array(0, 10)
    bob_process1.shared_mem.init_new_array(0, 10)
    bob_memmgr.add_process(bob_process0)
    bob_memmgr.add_process(bob_process1)
    bob = BobProtocol("bob", bob_processor, {0: bob_process0, 1: bob_process1})
    bob_processor._interface.start()  # also starts peer listeners
    bob.start()
    bob_egp.start()

    link_prot = alice_egp._ll_prot  # same as bob_egp._ll_prot
    link_prot.start()

    assert (
        alice.processes[0].shared_mem.get_array_part(
            result_array_addr,
            SER_RESPONSE_KEEP_LEN * 0 + SER_RESPONSE_KEEP_IDX_GOODNESS,
        )
        is None
    )

    ns.sim_run()

    assert alice.memmgr.phys_id_for(alice.processes[0].pid, 0) == 0
    assert alice.memmgr.phys_id_for(alice.processes[1].pid, 0) == 1
    assert bob.memmgr.phys_id_for(bob.processes[0].pid, 0) == 0
    assert bob.memmgr.phys_id_for(bob.processes[1].pid, 0) == 1

    assert (
        alice.processes[0].shared_mem.get_array_part(
            result_array_addr,
            SER_RESPONSE_KEEP_LEN * 0 + SER_RESPONSE_KEEP_IDX_GOODNESS,
        )
        is not None
    )

    alice_qubit = alice_processor.qdevice.get_local_qubit(0)
    bob_qubit = bob_processor.qdevice.get_local_qubit(0)
    assert has_multi_state([alice_qubit, bob_qubit], ketstates.b00)

    alice_qubit = alice_processor.qdevice.get_local_qubit(1)
    bob_qubit = bob_processor.qdevice.get_local_qubit(1)
    assert has_multi_state([alice_qubit, bob_qubit], ketstates.b00)

    alice_qubit = alice_processor.qdevice.get_local_qubit(2)
    assert alice_qubit is None
    bob_qubit = bob_processor.qdevice.get_local_qubit(2)
    assert bob_qubit is None


def test_4():
    alice_processor, bob_processor = setup_components_generic(num_qubits=3)

    topology = Topology(comm_ids={0, 1, 2}, mem_ids={0, 1, 2})
    unit_module = UnitModule.from_topology(topology)

    alice_node = alice_processor._interface._comp.node
    bob_node = bob_processor._interface._comp.node

    result_array_addr = 0

    alice_requests = {
        0: create_netstack_create_request(remote_id=bob_node.ID),
        1: create_netstack_create_request(remote_id=bob_node.ID),
    }
    bob_requests = {
        0: create_netstack_receive_request(remote_id=alice_node.ID),
        1: create_netstack_receive_request(remote_id=alice_node.ID),
    }

    alice_memmgr = alice_processor._interface.memmgr
    bob_memmgr = bob_processor._interface.memmgr

    alice_egpmgr = alice_processor._interface.egpmgr
    bob_egpmgr = bob_processor._interface.egpmgr

    alice_egp, bob_egp = create_egp_protocols(alice_node, bob_node)
    alice_egpmgr.add_egp(bob_node.ID, alice_egp)
    bob_egpmgr.add_egp(alice_node.ID, bob_egp)

    class NetstackProcessorProtocol(Protocol):
        def __init__(
            self,
            name: str,
            processor: NetstackProcessor,
            processes: Dict[int, IqoalaProcess],
        ) -> None:
            super().__init__(name)
            self._processor = processor
            self._memmgr = processor._interface.memmgr
            self._processes = processes

        @property
        def processes(self) -> Dict[int, IqoalaProcess]:
            return self._processes

        @property
        def memmgr(self) -> MemoryManager:
            return self._memmgr

    class AliceProtocol(NetstackProcessorProtocol):
        def run(self) -> Generator[EventExpression, None, None]:
            for pid, process in self.processes.items():
                yield from self._processor.handle_create_request(
                    process, alice_requests[pid]
                )

    class BobProtocol(NetstackProcessorProtocol):
        def run(self) -> Generator[EventExpression, None, None]:
            for pid, process in self.processes.items():
                yield from self._processor.handle_receive_request(
                    process, bob_requests[pid]
                )

    alice_process0 = create_process(pid=0, unit_module=unit_module)
    alice_process1 = create_process(pid=1, unit_module=unit_module)
    alice_process0.shared_mem.init_new_array(0, 10)
    alice_process1.shared_mem.init_new_array(0, 10)
    alice_memmgr.add_process(alice_process0)
    alice_memmgr.add_process(alice_process1)
    alice = AliceProtocol(
        "alice", alice_processor, {0: alice_process0, 1: alice_process1}
    )
    alice_processor._interface.start()  # also starts peer listeners
    alice.start()
    alice_egp.start()

    bob_process0 = create_process(pid=0, unit_module=unit_module)
    bob_process1 = create_process(pid=1, unit_module=unit_module)
    bob_process0.shared_mem.init_new_array(0, 10)
    bob_process1.shared_mem.init_new_array(0, 10)
    bob_memmgr.add_process(bob_process0)
    bob_memmgr.add_process(bob_process1)
    bob = BobProtocol("bob", bob_processor, {0: bob_process0, 1: bob_process1})
    bob_processor._interface.start()  # also starts peer listeners
    bob.start()
    bob_egp.start()

    link_prot = alice_egp._ll_prot  # same as bob_egp._ll_prot
    link_prot.start()

    assert (
        alice.processes[0].shared_mem.get_array_part(
            result_array_addr,
            SER_RESPONSE_KEEP_LEN * 0 + SER_RESPONSE_KEEP_IDX_GOODNESS,
        )
        is None
    )

    ns.sim_run()

    assert alice.memmgr.phys_id_for(alice.processes[0].pid, 0) == 0
    assert alice.memmgr.phys_id_for(alice.processes[1].pid, 0) == 1
    assert bob.memmgr.phys_id_for(bob.processes[0].pid, 0) == 0
    assert bob.memmgr.phys_id_for(bob.processes[1].pid, 0) == 1

    assert (
        alice.processes[0].shared_mem.get_array_part(
            result_array_addr,
            SER_RESPONSE_KEEP_LEN * 0 + SER_RESPONSE_KEEP_IDX_GOODNESS,
        )
        is not None
    )

    alice_qubit = alice_processor.qdevice.get_local_qubit(0)
    bob_qubit = bob_processor.qdevice.get_local_qubit(0)
    assert has_multi_state([alice_qubit, bob_qubit], ketstates.b00)

    alice_qubit = alice_processor.qdevice.get_local_qubit(1)
    bob_qubit = bob_processor.qdevice.get_local_qubit(1)
    assert has_multi_state([alice_qubit, bob_qubit], ketstates.b00)

    alice_qubit = alice_processor.qdevice.get_local_qubit(2)
    assert alice_qubit is None
    bob_qubit = bob_processor.qdevice.get_local_qubit(2)
    assert bob_qubit is None


if __name__ == "__main__":
    test_1()
    test_2()
    test_3()
    test_4()



from typing import Dict

from qoala.sim.egp import EgpProtocol


class NoEgpError(Exception):
    pass


class EgpManager:
    def __init__(self) -> None:
        self._egps: Dict[int, EgpProtocol] = {}

    def add_egp(self, remote_id: int, prot: EgpProtocol) -> None:
        self._egps[remote_id] = prot

    def get_egp(self, remote_id: int) -> EgpProtocol:
        if remote_id not in self._egps:
            raise NoEgpError
        return self._egps[remote_id]

    @property
    def egps(self) -> Dict[int, EgpProtocol]:
        return self._egps



from __future__ import annotations

from qoala.runtime.environment import GlobalEnvironment, GlobalNodeInfo
from qoala.sim.procnodecomp import ProcNodeComponent


def create_procnodecomp(num_other_nodes: int) -> ProcNodeComponent:
    env = GlobalEnvironment()

    node_info = GlobalNodeInfo("alice", 0)
    env.add_node(0, node_info)

    for id in range(1, num_other_nodes + 1):
        node_info = GlobalNodeInfo(f"node_{id}", id)
        env.add_node(id, node_info)

    return ProcNodeComponent(name="alice", qprocessor=None, global_env=env)


def test_no_other_nodes():
    comp = create_procnodecomp(num_other_nodes=0)

    # should not have any ports
    assert len(comp.ports) == 0


def test_one_other_node():
    comp = create_procnodecomp(num_other_nodes=1)

    # should have 2 host peer ports + 2 netstack peer ports
    assert len(comp.ports) == 4
    assert "host_peer_node_1_in" in comp.ports
    assert "host_peer_node_1_out" in comp.ports
    assert "netstack_peer_node_1_in" in comp.ports
    assert "netstack_peer_node_1_out" in comp.ports

    # Test properties
    assert comp.host_peer_in_port("node_1") == comp.ports["host_peer_node_1_in"]
    assert comp.host_peer_out_port("node_1") == comp.ports["host_peer_node_1_out"]
    assert comp.netstack_peer_in_port("node_1") == comp.ports["netstack_peer_node_1_in"]
    assert (
        comp.netstack_peer_out_port("node_1") == comp.ports["netstack_peer_node_1_out"]
    )


def test_many_other_nodes():
    comp = create_procnodecomp(num_other_nodes=5)

    # should 5 * 4 peer ports
    assert len(comp.ports) == 20

    for i in range(1, 6):
        assert f"host_peer_node_{i}_in" in comp.ports
        assert f"host_peer_node_{i}_out" in comp.ports
        assert f"netstack_peer_node_{i}_in" in comp.ports
        assert f"netstack_peer_node_{i}_out" in comp.ports
        # Test properties
        assert (
            comp.host_peer_in_port(f"node_{i}") == comp.ports[f"host_peer_node_{i}_in"]
        )
        assert (
            comp.host_peer_out_port(f"node_{i}")
            == comp.ports[f"host_peer_node_{i}_out"]
        )
        assert (
            comp.netstack_peer_in_port(f"node_{i}")
            == comp.ports[f"netstack_peer_node_{i}_in"]
        )
        assert (
            comp.netstack_peer_out_port(f"node_{i}")
            == comp.ports[f"netstack_peer_node_{i}_out"]
        )


if __name__ == "__main__":
    test_no_other_nodes()
    test_one_other_node()
    test_many_other_nodes()



from __future__ import annotations

import logging
from typing import Generator, Optional, Union

import netsquid as ns
from netqasm.lang.instr import NetQASMInstruction, core, nv, vanilla
from netqasm.lang.operand import Register
from netsquid.components.instructions import (
    INSTR_CNOT,
    INSTR_CXDIR,
    INSTR_CYDIR,
    INSTR_CZ,
    INSTR_H,
    INSTR_INIT,
    INSTR_MEASURE,
    INSTR_ROT_X,
    INSTR_ROT_Y,
    INSTR_ROT_Z,
    INSTR_X,
    INSTR_Y,
    INSTR_Z,
)
from netsquid.components.instructions import Instruction as NsInstr
from netsquid.qubits import qubitapi

from pydynaa import EventExpression
from qoala.sim.constants import PI, PI_OVER_2
from qoala.sim.globals import GlobalSimData
from qoala.sim.logging import LogManager
from qoala.sim.memmgr import NotAllocatedError
from qoala.sim.memory import ProgramMemory
from qoala.sim.message import Message
from qoala.sim.process import IqoalaProcess
from qoala.sim.qdevice import QDevice, QDeviceCommand
from qoala.sim.qnosinterface import QnosInterface
from qoala.sim.requests import (
    NetstackBreakpointCreateRequest,
    NetstackBreakpointReceiveRequest,
)


class QnosProcessor:
    """Does not have state itself."""

    def __init__(self, interface: QnosInterface, asynchronous: bool = False) -> None:
        self._interface = interface
        self._asynchronous = asynchronous

        # TODO: add way to set instr latency through constructor
        self._instr_latency: float = 0.0

        # TODO: rewrite
        self._name = f"{interface.name}_QnosProcessor"

        self._logger: logging.Logger = LogManager.get_stack_logger(  # type: ignore
            f"{self.__class__.__name__}({self._name})"
        )

        # memory of current program, only not-None when processor is active
        self._current_prog_mem: Optional[ProgramMemory] = None

    def _prog_mem(self) -> ProgramMemory:
        # May only be called when processor is active
        assert self._current_prog_mem is not None
        return self._current_prog_mem

    @property
    def qdevice(self) -> QDevice:
        return self._interface.qdevice

    @property
    def instr_latency(self) -> float:
        return self._instr_latency

    @instr_latency.setter
    def instr_latency(self, latency: float) -> None:
        self._instr_latency = latency

    def assign(
        self, process: IqoalaProcess, subrt_name: str, instr_idx: int
    ) -> Generator[EventExpression, None, int]:
        iqoala_subrt = process.subroutines[subrt_name]
        pid = process.prog_instance.pid

        self._current_prog_mem = process.prog_memory

        subroutine = iqoala_subrt.subroutine

        # TODO: handle program counter and jumping!!
        instr = subroutine.instructions[instr_idx]
        self._logger.debug(
            f"{ns.sim_time()} interpreting instruction {instr} at line {instr_idx}"
        )

        next_instr_idx: int

        if (
            isinstance(instr, core.JmpInstruction)
            or isinstance(instr, core.BranchUnaryInstruction)
            or isinstance(instr, core.BranchBinaryInstruction)
        ):
            if new_line := self._interpret_branch_instr(pid, instr):
                next_instr_idx = new_line
            else:
                next_instr_idx = instr_idx + 1
        else:
            generator = self._interpret_instruction(pid, instr)
            if generator:
                yield from generator
            next_instr_idx = instr_idx + 1

        self._current_prog_mem = None
        return next_instr_idx

    def _interpret_instruction(
        self, pid: int, instr: NetQASMInstruction
    ) -> Optional[Generator[EventExpression, None, None]]:
        if isinstance(instr, core.SetInstruction):
            return self._interpret_set(pid, instr)
        elif isinstance(instr, core.QAllocInstruction):
            return self._interpret_qalloc(pid, instr)
        elif isinstance(instr, core.QFreeInstruction):
            return self._interpret_qfree(pid, instr)
        elif isinstance(instr, core.StoreInstruction):
            return self._interpret_store(pid, instr)
        elif isinstance(instr, core.LoadInstruction):
            return self._interpret_load(pid, instr)
        elif isinstance(instr, core.LeaInstruction):
            return self._interpret_lea(pid, instr)
        elif isinstance(instr, core.UndefInstruction):
            return self._interpret_undef(pid, instr)
        elif isinstance(instr, core.ArrayInstruction):
            return self._interpret_array(pid, instr)
        elif isinstance(instr, core.InitInstruction):
            return self._interpret_init(pid, instr)
        elif isinstance(instr, core.MeasInstruction):
            return self._interpret_meas(pid, instr)
        elif isinstance(instr, core.CreateEPRInstruction):
            return self._interpret_create_epr(pid, instr)
        elif isinstance(instr, core.RecvEPRInstruction):
            return self._interpret_recv_epr(pid, instr)
        elif isinstance(instr, core.WaitAllInstruction):
            return self._interpret_wait_all(pid, instr)
        elif isinstance(instr, core.RetRegInstruction):
            return None
        elif isinstance(instr, core.RetArrInstruction):
            return None
        elif isinstance(instr, core.SingleQubitInstruction):
            return self._interpret_single_qubit_instr(pid, instr)
        elif isinstance(instr, core.TwoQubitInstruction):
            return self._interpret_two_qubit_instr(pid, instr)
        elif isinstance(instr, core.RotationInstruction):
            return self._interpret_single_rotation_instr(pid, instr)
        elif isinstance(instr, core.ControlledRotationInstruction):
            return self._interpret_controlled_rotation_instr(pid, instr)
        elif isinstance(instr, core.ClassicalOpInstruction) or isinstance(
            instr, core.ClassicalOpModInstruction
        ):
            return self._interpret_binary_classical_instr(pid, instr)
        elif isinstance(instr, core.BreakpointInstruction):
            return self._interpret_breakpoint(pid, instr)
        else:
            raise RuntimeError(f"Invalid instruction {instr}")

    def _interpret_breakpoint(
        self, pid: int, instr: core.BreakpointInstruction
    ) -> Optional[Generator[EventExpression, None, None]]:
        if instr.action.value == 0:
            self._logger.info("BREAKPOINT: no action taken")
        elif instr.action.value == 1:
            self._logger.info("BREAKPOINT: dumping local state:")
            for i in range(self.qdevice.num_positions):
                if self.qdevice.mem_positions[i].in_use:
                    q = self.qdevice.peek(i, skip_noise=True)
                    qstate = qubitapi.reduced_dm(q)
                    self._logger.info(f"physical qubit {i}:\n{qstate}")

            GlobalSimData.get_quantum_state(save=True)  # TODO: rewrite this
        elif instr.action.value == 2:
            self._logger.info("BREAKPOINT: dumping global state:")
            if instr.role.value == 0:
                self._interface.send_netstack_msg(
                    Message(content=NetstackBreakpointCreateRequest(pid))
                )
                ready = yield from self._interface.receive_netstack_msg()
                assert ready.content == "breakpoint ready"

                state = GlobalSimData.get_quantum_state(save=True)
                self._logger.info(state)

                self._interface.send_netstack_msg(Message(content="breakpoint end"))
                finished = yield from self._interface.receive_netstack_msg()
                assert finished.content == "breakpoint finished"
            elif instr.role.value == 1:
                self._interface.send_netstack_msg(
                    Message(content=NetstackBreakpointReceiveRequest(pid))
                )
                ready = yield from self._interface.receive_netstack_msg()
                assert ready.content == "breakpoint ready"
                self._interface.send_netstack_msg(Message(content="breakpoint end"))
                finished = yield from self._interface.receive_netstack_msg()
                assert finished.content == "breakpoint finished"
            else:
                raise ValueError
        else:
            raise ValueError

    def _interpret_set(
        self, pid: int, instr: core.SetInstruction
    ) -> Optional[Generator[EventExpression, None, None]]:
        self._logger.debug(f"Set register {instr.reg} to {instr.imm}")
        shared_mem = self._prog_mem().shared_mem
        shared_mem.set_reg_value(instr.reg, instr.imm.value)
        yield from self._interface.wait(self._instr_latency)
        return None

    def _interpret_qalloc(
        self, pid: int, instr: core.QAllocInstruction
    ) -> Optional[Generator[EventExpression, None, None]]:
        shared_mem = self._prog_mem().shared_mem

        virt_id = shared_mem.get_reg_value(instr.reg)
        if virt_id is None:
            raise RuntimeError(f"qubit address in register {instr.reg} is not defined")
        self._logger.debug(f"Allocating qubit with virtual ID {virt_id}")
        self._interface.memmgr.allocate(pid, virt_id)

        return None

    def _interpret_qfree(
        self, pid: int, instr: core.QFreeInstruction
    ) -> Optional[Generator[EventExpression, None, None]]:
        shared_mem = self._prog_mem().shared_mem

        virt_id = shared_mem.get_reg_value(instr.reg)
        assert virt_id is not None
        self._logger.debug(f"Freeing virtual qubit {virt_id}")
        self._interface.memmgr.free(pid, virt_id)
        self._interface.signal_memory_freed()

        return None

    def _interpret_store(
        self, pid: int, instr: core.StoreInstruction
    ) -> Optional[Generator[EventExpression, None, None]]:
        shared_mem = self._prog_mem().shared_mem

        value = shared_mem.get_reg_value(instr.reg)
        if value is None:
            raise RuntimeError(f"value in register {instr.reg} is not defined")
        self._logger.debug(
            f"Storing value {value} from register {instr.reg} "
            f"to array entry {instr.entry}"
        )

        shared_mem.set_array_entry(instr.entry, value)
        return None

    def _interpret_load(
        self, pid: int, instr: core.LoadInstruction
    ) -> Optional[Generator[EventExpression, None, None]]:
        shared_mem = self._prog_mem().shared_mem

        value = shared_mem.get_array_entry(instr.entry)
        if value is None:
            raise RuntimeError(f"array value at {instr.entry} is not defined")
        self._logger.debug(
            f"Storing value {value} from array entry {instr.entry} "
            f"to register {instr.reg}"
        )

        shared_mem.set_reg_value(instr.reg, value)
        return None

    def _interpret_lea(
        self, pid: int, instr: core.LeaInstruction
    ) -> Optional[Generator[EventExpression, None, None]]:
        shared_mem = self._prog_mem().shared_mem
        self._logger.debug(
            f"Storing address of {instr.address} to register {instr.reg}"
        )
        shared_mem.set_reg_value(instr.reg, instr.address.address)
        return None

    def _interpret_undef(
        self, pid: int, instr: core.UndefInstruction
    ) -> Optional[Generator[EventExpression, None, None]]:
        shared_mem = self._prog_mem().shared_mem
        self._logger.debug(f"Unset array entry {instr.entry}")
        shared_mem.set_array_entry(instr.entry, None)
        return None

    def _interpret_array(
        self, pid: int, instr: core.ArrayInstruction
    ) -> Optional[Generator[EventExpression, None, None]]:
        shared_mem = self._prog_mem().shared_mem

        length = shared_mem.get_reg_value(instr.size)
        assert length is not None
        self._logger.debug(
            f"Initializing an array of length {length} at address {instr.address}"
        )

        shared_mem.init_new_array(instr.address.address, length)
        return None

    def _interpret_branch_instr(
        self,
        pid: int,
        instr: Union[
            core.BranchUnaryInstruction,
            core.BranchBinaryInstruction,
            core.JmpInstruction,
        ],
    ) -> Optional[int]:
        """Returns line to jump to, or None if no jump happens."""
        shared_mem = self._prog_mem().shared_mem
        a, b = None, None
        registers = []
        if isinstance(instr, core.BranchUnaryInstruction):
            a = shared_mem.get_reg_value(instr.reg)
            registers = [instr.reg]
        elif isinstance(instr, core.BranchBinaryInstruction):
            a = shared_mem.get_reg_value(instr.reg0)
            b = shared_mem.get_reg_value(instr.reg1)
            registers = [instr.reg0, instr.reg1]

        if isinstance(instr, core.JmpInstruction):
            condition = True
        elif isinstance(instr, core.BranchUnaryInstruction):
            condition = instr.check_condition(a)
        elif isinstance(instr, core.BranchBinaryInstruction):
            condition = instr.check_condition(a, b)

        if condition:
            jump_address = instr.line
            self._logger.debug(
                f"Branching to line {jump_address}, since {instr}(a={a}, b={b}) "
                f"is True, with values from registers {registers}"
            )
            return jump_address.value  # type: ignore
        else:
            self._logger.debug(
                f"Don't branch, since {instr}(a={a}, b={b}) "
                f"is False, with values from registers {registers}"
            )
            return None

    def _interpret_binary_classical_instr(
        self,
        pid: int,
        instr: Union[
            core.ClassicalOpInstruction,
            core.ClassicalOpModInstruction,
        ],
    ) -> Optional[Generator[EventExpression, None, None]]:
        shared_mem = self._prog_mem().shared_mem
        mod = None
        if isinstance(instr, core.ClassicalOpModInstruction):
            mod = shared_mem.get_reg_value(instr.regmod)
        if mod is not None and mod < 1:
            raise RuntimeError(f"Modulus needs to be greater or equal to 1, not {mod}")
        a = shared_mem.get_reg_value(instr.regin0)
        b = shared_mem.get_reg_value(instr.regin1)
        assert a is not None
        assert b is not None
        value = self._compute_binary_classical_instr(instr, a, b, mod=mod)
        mod_str = "" if mod is None else f"(mod {mod})"
        self._logger.debug(
            f"Performing {instr} of a={a} and b={b} {mod_str} "
            f"and storing the value {value} at register {instr.regout}"
        )
        shared_mem.set_reg_value(instr.regout, value)
        return None

    def _compute_binary_classical_instr(
        self, instr: NetQASMInstruction, a: int, b: int, mod: Optional[int] = 1
    ) -> int:
        if isinstance(instr, core.AddInstruction):
            return a + b
        elif isinstance(instr, core.AddmInstruction):
            assert mod is not None
            return (a + b) % mod
        elif isinstance(instr, core.SubInstruction):
            return a - b
        elif isinstance(instr, core.SubmInstruction):
            assert mod is not None
            return (a - b) % mod
        else:
            raise ValueError(f"{instr} cannot be used as binary classical function")

    def _interpret_init(
        self, pid: int, instr: core.InitInstruction
    ) -> Generator[EventExpression, None, None]:
        raise NotImplementedError

    def _do_single_rotation(
        self,
        pid: int,
        instr: core.RotationInstruction,
        ns_instr: NsInstr,
    ) -> Generator[EventExpression, None, None]:
        shared_mem = self._prog_mem().shared_mem
        virt_id = shared_mem.get_reg_value(instr.reg)
        phys_id = self._interface.memmgr.phys_id_for(pid, virt_id)
        if phys_id is None:
            raise NotAllocatedError
        angle = self._get_rotation_angle_from_operands(
            n=instr.angle_num.value,
            d=instr.angle_denom.value,
        )
        self._logger.debug(
            f"Performing {instr} with angle {angle} on virtual qubit "
            f"{virt_id} (physical ID: {phys_id})"
        )
        commands = [QDeviceCommand(ns_instr, [phys_id], angle=angle)]
        yield from self.qdevice.execute_commands(commands)
        return None

    def _interpret_single_rotation_instr(
        self, pid: int, instr: nv.RotXInstruction
    ) -> Generator[EventExpression, None, None]:
        raise NotImplementedError

    def _do_controlled_rotation(
        self,
        pid: int,
        instr: core.ControlledRotationInstruction,
        ns_instr: NsInstr,
    ) -> Generator[EventExpression, None, None]:
        shared_mem = self._prog_mem().shared_mem
        virt_id0 = shared_mem.get_reg_value(instr.reg0)
        phys_id0 = self._interface.memmgr.phys_id_for(pid, virt_id0)
        if phys_id0 is None:
            raise NotAllocatedError
        virt_id1 = shared_mem.get_reg_value(instr.reg1)
        phys_id1 = self._interface.memmgr.phys_id_for(pid, virt_id1)
        if phys_id1 is None:
            raise NotAllocatedError
        angle = self._get_rotation_angle_from_operands(
            n=instr.angle_num.value,
            d=instr.angle_denom.value,
        )
        self._logger.debug(
            f"Performing {instr} with angle {angle} on virtual qubits "
            f"{virt_id0} and {virt_id1} (physical IDs: {phys_id0} and {phys_id1})"
        )
        commands = [QDeviceCommand(ns_instr, [phys_id0, phys_id1], angle=angle)]
        yield from self.qdevice.execute_commands(commands)
        return None

    def _interpret_controlled_rotation_instr(
        self, pid: int, instr: core.ControlledRotationInstruction
    ) -> Generator[EventExpression, None, None]:
        raise NotImplementedError

    def _get_rotation_angle_from_operands(self, n: int, d: int) -> float:
        return float(n * PI / (2**d))

    def _interpret_meas(
        self, pid: int, instr: core.MeasInstruction
    ) -> Generator[EventExpression, None, None]:
        raise NotImplementedError

    def _interpret_create_epr(
        self, pid: int, instr: core.CreateEPRInstruction
    ) -> Optional[Generator[EventExpression, None, None]]:
        # shared_mem = self._prog_mem().shared_mem
        # remote_node_id = shared_mem.get_reg_value(instr.remote_node_id)
        # epr_socket_id = shared_mem.get_reg_value(instr.epr_socket_id)
        # qubit_array_addr = shared_mem.get_reg_value(instr.qubit_addr_array)
        # arg_array_addr = shared_mem.get_reg_value(instr.arg_array)
        # result_array_addr = shared_mem.get_reg_value(instr.ent_results_array)
        # assert remote_node_id is not None
        # assert epr_socket_id is not None
        # # qubit_array_addr can be None
        # assert arg_array_addr is not None
        # assert result_array_addr is not None
        # self._logger.debug(
        #     f"Creating EPR pair with remote node id {remote_node_id} "
        #     f"and EPR socket ID {epr_socket_id}, "
        #     f"using qubit addresses stored in array with address {qubit_array_addr}, "
        #     f"using arguments stored in array with address {arg_array_addr}, "
        #     f"placing the entanglement information in array at "
        #     f"address {result_array_addr}"
        # )

        # raw_args = shared_mem.get_array(arg_array_addr)
        # raw_typ = raw_args[SER_CREATE_IDX_TYPE]
        # typ = {
        #     0: EprCreateType.CREATE_KEEP,
        #     1: EprCreateType.MEASURE_DIRECTLY,
        #     2: EprCreateType.REMOTE_STATE_PREP,
        # }[raw_typ]
        # num_pairs = raw_args[SER_CREATE_IDX_NUMBER]

        # if qubit_array_addr is not None:
        #     virt_ids = shared_mem.get_array(qubit_array_addr)
        # else:
        #     virt_ids = [0 for _ in range(num_pairs)]

        # # TODO: get fidelity from EPR socket

        # request = NetstackCreateRequest(
        #     remote_id=remote_node_id,
        #     epr_socket_id=epr_socket_id,
        #     typ=typ,
        #     num_pairs=num_pairs,
        #     fidelity=1.0,  # for now, always just ask for the best
        #     virt_qubit_ids=virt_ids,
        #     result_array_addr=result_array_addr,
        # )

        # if self._asynchronous:
        #     self._interface.send_netstack_msg(Message(content=request))
        #     # result = yield from self._interface.receive_netstack_msg()
        #     # self._logger.debug(f"result from netstack: {result}")
        # else:
        #     self._prog_mem().requests = [request]
        return None

    def _interpret_recv_epr(
        self, pid: int, instr: core.RecvEPRInstruction
    ) -> Optional[Generator[EventExpression, None, None]]:
        # shared_mem = self._prog_mem().shared_mem
        # remote_node_id = shared_mem.get_reg_value(instr.remote_node_id)
        # epr_socket_id = shared_mem.get_reg_value(instr.epr_socket_id)
        # qubit_array_addr = shared_mem.get_reg_value(instr.qubit_addr_array)
        # result_array_addr = shared_mem.get_reg_value(instr.ent_results_array)
        # assert remote_node_id is not None
        # assert epr_socket_id is not None
        # # qubit_array_addr can be None
        # assert result_array_addr is not None
        # self._logger.debug(
        #     f"Receiving EPR pair with remote node id {remote_node_id} "
        #     f"and EPR socket ID {epr_socket_id}, "
        #     f"using qubit addresses stored in array with address {qubit_array_addr}, "
        #     f"placing the entanglement information in array at "
        #     f"address {result_array_addr}"
        # )

        # # We don't allow None for qubit_array_addr since creating the array manually
        # # requires the number of pairs (for the length), which we don't know.
        # assert qubit_array_addr is not None
        # virt_ids = shared_mem.get_array(qubit_array_addr)

        # request = NetstackReceiveRequest(
        #     remote_id=remote_node_id,
        #     epr_socket_id=epr_socket_id,
        #     typ=None,  # TODO: fix recv_epr instruction
        #     num_pairs=None,  # TODO: fix recv_epr instruction
        #     fidelity=1.0,  # for now, always just ask for the best
        #     virt_qubit_ids=virt_ids,
        #     result_array_addr=result_array_addr,
        # )

        # if self._asynchronous:
        #     self._interface.send_netstack_msg(Message(content=request))
        #     # result = yield from self._interface.receive_netstack_msg()
        #     # self._logger.debug(f"result from netstack: {result}")
        # else:
        #     self._prog_mem().requests = [request]
        return None

    def _interpret_wait_all(
        self, pid: int, instr: core.WaitAllInstruction
    ) -> Generator[EventExpression, None, None]:
        shared_mem = self._prog_mem().shared_mem
        self._logger.debug(
            f"Waiting for all entries in array slice {instr.slice} to become defined"
        )
        assert isinstance(instr.slice.start, Register)
        assert isinstance(instr.slice.stop, Register)
        start: int = shared_mem.get_reg_value(instr.slice.start)
        end: int = shared_mem.get_reg_value(instr.slice.stop)
        addr: int = instr.slice.address.address

        self._logger.debug(
            f"checking if @{addr}[{start}:{end}] has values for app ID {pid}"
        )

        while True:
            values = shared_mem.get_array_values(addr, start, end)
            if any(v is None for v in values):
                self._logger.debug(
                    f"waiting for netstack to write to @{addr}[{start}:{end}] "
                    f"for app ID {pid}"
                )
                yield from self._interface.receive_netstack_msg()
                self._logger.debug("netstack wrote something")
            else:
                break
        self._interface.flush_netstack_msgs()
        self._logger.debug("all entries were written")

        self._logger.info(f"\nFinished waiting for array slice {instr.slice}")
        return None

    def _interpret_ret_reg(
        self, pid: int, instr: core.RetRegInstruction
    ) -> Optional[Generator[EventExpression, None, None]]:
        return None

    def _interpret_ret_arr(
        self, pid: int, instr: core.RetArrInstruction
    ) -> Optional[Generator[EventExpression, None, None]]:
        return None

    def _interpret_single_qubit_instr(
        self, pid: int, instr: core.SingleQubitInstruction
    ) -> Generator[EventExpression, None, None]:
        raise NotImplementedError

    def _interpret_two_qubit_instr(
        self, pid: int, instr: core.SingleQubitInstruction
    ) -> Generator[EventExpression, None, None]:
        raise NotImplementedError


class GenericProcessor(QnosProcessor):
    """A `Processor` for nodes with a generic quantum hardware."""

    def _interpret_init(
        self, pid: int, instr: core.InitInstruction
    ) -> Generator[EventExpression, None, None]:
        shared_mem = self._prog_mem().shared_mem
        virt_id = shared_mem.get_reg_value(instr.reg)
        phys_id = self._interface.memmgr.phys_id_for(pid, virt_id)
        if phys_id is None:
            raise NotAllocatedError
        self._logger.debug(
            f"Performing {instr} on virtual qubit "
            f"{virt_id} (physical ID: {phys_id})"
        )
        commands = [QDeviceCommand(INSTR_INIT, [phys_id])]
        yield from self.qdevice.execute_commands(commands)
        return None

    def _interpret_meas(
        self, pid: int, instr: core.MeasInstruction
    ) -> Generator[EventExpression, None, None]:
        shared_mem = self._prog_mem().shared_mem
        virt_id = shared_mem.get_reg_value(instr.qreg)
        phys_id = self._interface.memmgr.phys_id_for(pid, virt_id)
        if phys_id is None:
            raise NotAllocatedError

        self._logger.debug(
            f"Measuring qubit {virt_id} (physical ID: {phys_id}), "
            f"placing the outcome in register {instr.creg}"
        )

        commands = [QDeviceCommand(INSTR_MEASURE, [phys_id])]
        outcome = yield from self.qdevice.execute_commands(commands)
        shared_mem.set_reg_value(instr.creg, outcome)
        return None

    def _interpret_single_qubit_instr(
        self, pid: int, instr: core.SingleQubitInstruction
    ) -> Generator[EventExpression, None, None]:
        shared_mem = self._prog_mem().shared_mem
        virt_id = shared_mem.get_reg_value(instr.qreg)
        phys_id = self._interface.memmgr.phys_id_for(pid, virt_id)
        if phys_id is None:
            raise NotAllocatedError
        if isinstance(instr, vanilla.GateXInstruction):
            commands = [QDeviceCommand(INSTR_X, [phys_id])]
            yield from self.qdevice.execute_commands(commands)
        elif isinstance(instr, vanilla.GateYInstruction):
            commands = [QDeviceCommand(INSTR_Y, [phys_id])]
            yield from self.qdevice.execute_commands(commands)
        elif isinstance(instr, vanilla.GateZInstruction):
            commands = [QDeviceCommand(INSTR_Z, [phys_id])]
            yield from self.qdevice.execute_commands(commands)
        elif isinstance(instr, vanilla.GateHInstruction):
            commands = [QDeviceCommand(INSTR_H, [phys_id])]
            yield from self.qdevice.execute_commands(commands)
        else:
            raise RuntimeError(f"Unsupported instruction {instr}")
        return None

    def _interpret_single_rotation_instr(
        self, pid: int, instr: nv.RotXInstruction
    ) -> Generator[EventExpression, None, None]:
        if isinstance(instr, vanilla.RotXInstruction):
            yield from self._do_single_rotation(pid, instr, INSTR_ROT_X)
        elif isinstance(instr, vanilla.RotYInstruction):
            yield from self._do_single_rotation(pid, instr, INSTR_ROT_Y)
        elif isinstance(instr, vanilla.RotZInstruction):
            yield from self._do_single_rotation(pid, instr, INSTR_ROT_Z)
        else:
            raise RuntimeError(f"Unsupported instruction {instr}")
        return None

    def _interpret_controlled_rotation_instr(
        self, pid: int, instr: core.ControlledRotationInstruction
    ) -> Generator[EventExpression, None, None]:
        raise RuntimeError(f"Unsupported instruction {instr}")

    def _interpret_two_qubit_instr(
        self, pid: int, instr: core.SingleQubitInstruction
    ) -> Generator[EventExpression, None, None]:
        shared_mem = self._prog_mem().shared_mem
        virt_id0 = shared_mem.get_reg_value(instr.reg0)
        phys_id0 = self._interface.memmgr.phys_id_for(pid, virt_id0)
        if phys_id0 is None:
            raise NotAllocatedError
        virt_id1 = shared_mem.get_reg_value(instr.reg1)
        phys_id1 = self._interface.memmgr.phys_id_for(pid, virt_id1)
        if phys_id1 is None:
            raise NotAllocatedError
        if isinstance(instr, vanilla.CnotInstruction):
            commands = [QDeviceCommand(INSTR_CNOT, [phys_id0, phys_id1])]
            yield from self.qdevice.execute_commands(commands)
        elif isinstance(instr, vanilla.CphaseInstruction):
            commands = [QDeviceCommand(INSTR_CZ, [phys_id0, phys_id1])]
            yield from self.qdevice.execute_commands(commands)
        else:
            raise RuntimeError(f"Unsupported instruction {instr}")
        return None


class NVProcessor(QnosProcessor):
    """A `Processor` for nodes with a NV hardware."""

    def _interpret_init(
        self, pid: int, instr: core.InitInstruction
    ) -> Generator[EventExpression, None, None]:
        shared_mem = self._prog_mem().shared_mem
        virt_id = shared_mem.get_reg_value(instr.reg)
        phys_id = self._interface.memmgr.phys_id_for(pid, virt_id)
        if phys_id is None:
            raise NotAllocatedError
        self._logger.debug(
            f"Performing {instr} on virtual qubit "
            f"{virt_id} (physical ID: {phys_id})"
        )
        commands = [QDeviceCommand(INSTR_INIT, [phys_id])]
        yield from self.qdevice.execute_commands(commands)
        return None

    def _measure_electron(self) -> Generator[EventExpression, None, int]:
        commands = [QDeviceCommand(INSTR_MEASURE, [0])]
        outcome = yield from self.qdevice.execute_commands(commands)
        assert outcome is not None
        return outcome  # type: ignore

    def _move_carbon_to_electron_for_measure(
        self, carbon_id: int
    ) -> Generator[EventExpression, None, None]:
        commands = [
            QDeviceCommand(INSTR_INIT, [0]),
            QDeviceCommand(INSTR_ROT_Y, [0], angle=PI_OVER_2),
            QDeviceCommand(INSTR_CYDIR, [0, carbon_id], angle=-PI_OVER_2),
            QDeviceCommand(INSTR_ROT_X, [0], angle=-PI_OVER_2),
            QDeviceCommand(INSTR_CXDIR, [0, carbon_id], angle=PI_OVER_2),
            QDeviceCommand(INSTR_ROT_Y, [0], angle=-PI_OVER_2),
        ]
        yield from self.qdevice.execute_commands(commands)
        return None

    def _move_carbon_to_electron(
        self, carbon_id: int
    ) -> Generator[EventExpression, None, None]:
        # TODO: CHECK SEQUENCE OF GATES!!!
        commands = [
            QDeviceCommand(INSTR_INIT, [0]),
            QDeviceCommand(INSTR_ROT_Y, [0], angle=PI_OVER_2),
            QDeviceCommand(INSTR_CYDIR, [0, carbon_id], angle=-PI_OVER_2),
            QDeviceCommand(INSTR_ROT_X, [0], angle=-PI_OVER_2),
            QDeviceCommand(INSTR_CXDIR, [0, carbon_id], angle=PI_OVER_2),
            QDeviceCommand(INSTR_ROT_Y, [0], angle=-PI_OVER_2),
        ]
        yield from self.qdevice.execute_commands(commands)
        return None

    def _move_electron_to_carbon(
        self, carbon_id: int
    ) -> Generator[EventExpression, None, None]:
        commands = [
            QDeviceCommand(INSTR_INIT, [carbon_id]),
            QDeviceCommand(INSTR_ROT_Y, [0], angle=PI_OVER_2),
            QDeviceCommand(INSTR_CYDIR, [0, carbon_id], angle=-PI_OVER_2),
            QDeviceCommand(INSTR_ROT_X, [0], angle=-PI_OVER_2),
            QDeviceCommand(INSTR_CXDIR, [0, carbon_id], angle=PI_OVER_2),
        ]
        yield from self.qdevice.execute_commands(commands)
        return None

    def _interpret_meas(
        self, pid: int, instr: core.MeasInstruction
    ) -> Generator[EventExpression, None, None]:
        shared_mem = self._prog_mem().shared_mem
        virt_id = shared_mem.get_reg_value(instr.qreg)
        phys_id = self._interface.memmgr.phys_id_for(pid, virt_id)

        # Only the electron (phys ID 0) can be measured.
        # Measuring any other physical qubit (i.e one of the carbons) requires
        # freeing up the electron and moving the target qubit to the electron first.

        self._logger.debug(
            f"Measuring qubit {virt_id} (physical ID: {phys_id}), "
            f"placing the outcome in register {instr.creg}"
        )

        memmgr = self._interface.memmgr

        if phys_id == 0:
            # Measuring a comm qubit. This can be done immediately.
            outcome = yield from self._measure_electron()
            shared_mem.set_reg_value(instr.creg, outcome)
            memmgr.free(pid, virt_id)
            return None
        # else:

        # We want to measure a mem qubit.
        # Move it to the comm qubit first.
        # Check if comm qubit is already used by this process:
        if memmgr.phys_id_for(pid, 0) is not None:
            # Comm qubit is already allocated. Try to move it to a free mem qubit.
            mem_virt_id = memmgr.get_unmapped_mem_qubit(pid)
            memmgr.allocate(pid, mem_virt_id)
            mem_phys_id = memmgr.phys_id_for(mem_virt_id)

            # Move (temporarily) state from comm qubit to mem qubit.
            yield from self._move_electron_to_carbon(mem_phys_id)

            # Move state from qubit-to-measure to comm qubit.
            yield from self._move_carbon_to_electron_for_measure(phys_id)
            memmgr.free(pid, virt_id)
            self._interface.signal_memory_freed()

            # Measure comm qubit (containing state of qubit-to-measure).
            outcome = yield from self._measure_electron()
            shared_mem.set_reg_value(instr.creg, outcome)

            # Move temporarily-moved state back to comm qubit.
            yield from self._move_carbon_to_electron(mem_phys_id)

            # Free mem qubit that was temporarily used.
            memmgr.free(mem_virt_id)
            self._interface.signal_memory_freed()
        else:  # comm qubit not in use.
            # Allocate comm qubit.
            memmgr.allocate(pid, 0)

            # Move state-to-measure to comm qubit.
            yield from self._move_carbon_to_electron_for_measure(phys_id)

            # Free qubit that contained state-to-measure.
            memmgr.free(pid, virt_id)
            self._interface.signal_memory_freed()

            # Measure comm qubit.
            outcome = yield from self._measure_electron()
            shared_mem.set_reg_value(instr.creg, outcome)

            # Free comm qubit.
            memmgr.free(pid, 0)
            self._interface.signal_memory_freed()
        return None

    def _interpret_single_rotation_instr(
        self, pid: int, instr: nv.RotXInstruction
    ) -> Generator[EventExpression, None, None]:
        if isinstance(instr, nv.RotXInstruction):
            yield from self._do_single_rotation(pid, instr, INSTR_ROT_X)
        elif isinstance(instr, nv.RotYInstruction):
            yield from self._do_single_rotation(pid, instr, INSTR_ROT_Y)
        elif isinstance(instr, nv.RotZInstruction):
            yield from self._do_single_rotation(pid, instr, INSTR_ROT_Z)
        else:
            raise RuntimeError(f"Unsupported instruction {instr}")

    def _interpret_controlled_rotation_instr(
        self, pid: int, instr: core.ControlledRotationInstruction
    ) -> Generator[EventExpression, None, None]:
        if isinstance(instr, nv.ControlledRotXInstruction):
            yield from self._do_controlled_rotation(pid, instr, INSTR_CXDIR)
        elif isinstance(instr, nv.ControlledRotYInstruction):
            yield from self._do_controlled_rotation(pid, instr, INSTR_CYDIR)
        else:
            raise RuntimeError(f"Unsupported instruction {instr}")



from __future__ import annotations

from typing import List, Set, Tuple

import pytest

from qoala.lang.iqoala import IqoalaProgram, ProgramMeta
from qoala.runtime.program import ProgramInput, ProgramInstance, ProgramResult
from qoala.runtime.schedule import ProgramTaskList
from qoala.sim.memmgr import AllocError, MemoryManager
from qoala.sim.memory import CommQubitTrait, MemQubitTrait, ProgramMemory, UnitModule
from qoala.sim.process import IqoalaProcess
from qoala.sim.qdevice import QDevice


class MockQDevice(QDevice):
    def __init__(self) -> None:
        pass

    @property
    def comm_qubit_ids(self) -> Set[int]:
        return {0}

    @property
    def mem_qubit_ids(self) -> Set[int]:
        return {1}

    def set_mem_pos_in_use(self, id: int, in_use: bool) -> None:
        pass


def create_process(pid: int, unit_module: UnitModule) -> IqoalaProcess:
    program = IqoalaProgram(
        instructions=[], subroutines={}, meta=ProgramMeta.empty("prog")
    )
    instance = ProgramInstance(
        pid=pid,
        program=program,
        inputs=ProgramInput({}),
        tasks=ProgramTaskList.empty(program),
    )
    mem = ProgramMemory(pid=pid, unit_module=unit_module)

    process = IqoalaProcess(
        prog_instance=instance,
        prog_memory=mem,
        csockets={},
        epr_sockets=program.meta.epr_sockets,
        subroutines=program.subroutines,
        requests={},
        result=ProgramResult(values={}),
    )
    return process


def setup_manager() -> Tuple[int, MemoryManager]:
    qdevice = MockQDevice()
    mgr = MemoryManager("alice", qdevice)

    um = UnitModule(
        qubit_ids=[0, 1],
        qubit_traits={0: [CommQubitTrait], 1: [MemQubitTrait]},
        gate_traits={},
    )

    process = create_process(0, um)
    mgr.add_process(process)
    pid = process.pid

    return pid, mgr


def setup_manager_multiple_processes(
    num_processes: int,
) -> Tuple[List[int], MemoryManager]:
    qdevice = MockQDevice()
    mgr = MemoryManager("alice", qdevice)

    um = UnitModule(
        qubit_ids=[0, 1],
        qubit_traits={0: [CommQubitTrait], 1: [MemQubitTrait]},
        gate_traits={},
    )

    pids: List[int] = []

    for i in range(num_processes):
        process = create_process(i, um)
        mgr.add_process(process)
        pids.append(process.pid)

    return pids, mgr


def test_alloc_free_0():
    pid, mgr = setup_manager()

    assert mgr.phys_id_for(pid, 0) is None
    assert mgr.phys_id_for(pid, 1) is None
    assert mgr.virt_id_for(pid, 0) is None
    assert mgr.virt_id_for(pid, 1) is None

    mgr.allocate(pid, 0)
    assert mgr.phys_id_for(pid, 0) == 0
    assert mgr.phys_id_for(pid, 1) is None
    assert mgr.virt_id_for(pid, 0) == 0
    assert mgr.virt_id_for(pid, 1) is None

    mgr.free(pid, 0)
    assert mgr.phys_id_for(pid, 0) is None
    assert mgr.phys_id_for(pid, 1) is None
    assert mgr.virt_id_for(pid, 0) is None
    assert mgr.virt_id_for(pid, 1) is None


def test_alloc_free_0_1():
    pid, mgr = setup_manager()

    assert mgr.phys_id_for(pid, 0) is None
    assert mgr.phys_id_for(pid, 1) is None
    assert mgr.virt_id_for(pid, 0) is None
    assert mgr.virt_id_for(pid, 1) is None

    mgr.allocate(pid, 0)
    mgr.allocate(pid, 1)
    assert mgr.phys_id_for(pid, 0) == 0
    assert mgr.phys_id_for(pid, 1) == 1
    assert mgr.virt_id_for(pid, 0) == 0
    assert mgr.virt_id_for(pid, 1) == 1

    mgr.free(pid, 0)
    assert mgr.phys_id_for(pid, 0) is None
    assert mgr.phys_id_for(pid, 1) == 1
    assert mgr.virt_id_for(pid, 0) is None
    assert mgr.virt_id_for(pid, 1) == 1


def test_alloc_non_existing():
    pid, mgr = setup_manager()

    with pytest.raises(AllocError):
        mgr.allocate(pid, 2)


def test_alloc_already_allocated():
    pid, mgr = setup_manager()

    with pytest.raises(AllocError):
        mgr.allocate(pid, 1)
        mgr.allocate(pid, 1)


def test_free_alreay_freed():
    pid, mgr = setup_manager()

    with pytest.raises(AllocError):
        mgr.free(pid, 0)


def test_get_unmapped_qubit():
    pid, mgr = setup_manager()

    assert mgr.get_unmapped_mem_qubit(pid) == 1
    mgr.allocate(pid, 0)
    assert mgr.get_unmapped_mem_qubit(pid) == 1
    mgr.allocate(pid, 1)
    with pytest.raises(AllocError):
        mgr.get_unmapped_mem_qubit(pid)
    mgr.free(pid, 1)
    assert mgr.get_unmapped_mem_qubit(pid) == 1


def test_alloc_multiple_processes():
    [pid0, pid1], mgr = setup_manager_multiple_processes(2)

    assert mgr.phys_id_for(pid0, 0) is None
    assert mgr.phys_id_for(pid0, 1) is None
    assert mgr.virt_id_for(pid0, 0) is None
    assert mgr.virt_id_for(pid0, 1) is None
    assert mgr.phys_id_for(pid1, 0) is None
    assert mgr.phys_id_for(pid1, 1) is None
    assert mgr.virt_id_for(pid1, 0) is None
    assert mgr.virt_id_for(pid1, 1) is None

    mgr.allocate(pid0, 0)
    # Should allocate phys ID 0 for virt ID 0 of pid0
    assert mgr.phys_id_for(pid0, 0) == 0
    assert mgr.virt_id_for(pid0, 0) == 0

    with pytest.raises(AllocError):
        # Should try to allocate phys ID 0, but it's not available
        mgr.allocate(pid1, 0)

    mgr.allocate(pid1, 1)
    # Should have mapping:
    #   phys ID 0 : (pid0, virt0)
    #   phys ID 1 : (pid1, virt1)
    assert mgr.phys_id_for(pid0, 0) == 0
    assert mgr.phys_id_for(pid0, 1) is None
    assert mgr.virt_id_for(pid0, 0) == 0
    assert mgr.virt_id_for(pid0, 1) is None
    assert mgr.phys_id_for(pid1, 0) is None
    assert mgr.phys_id_for(pid1, 1) == 1
    assert mgr.virt_id_for(pid1, 0) is None
    assert mgr.virt_id_for(pid1, 1) == 1


if __name__ == "__main__":
    test_alloc_free_0()
    test_alloc_free_0_1()
    test_alloc_non_existing()
    test_alloc_already_allocated()
    test_free_alreay_freed()
    test_get_unmapped_qubit()
    test_alloc_multiple_processes()



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, Generator, List, Optional

from netqasm.lang.parsing import parse_text_subroutine
from netqasm.lang.subroutine import Subroutine

from pydynaa import EventExpression
from qoala.lang.iqoala import (
    AddCValueOp,
    AssignCValueOp,
    BitConditionalMultiplyConstantCValueOp,
    ClassicalIqoalaOp,
    IqoalaProgram,
    IqoalaSharedMemLoc,
    IqoalaSubroutine,
    IqoalaVector,
    MultiplyConstantCValueOp,
    ProgramMeta,
    ReceiveCMsgOp,
    ReturnResultOp,
    RunSubroutineOp,
    SendCMsgOp,
)
from qoala.runtime.program import ProgramInput, ProgramInstance, ProgramResult
from qoala.runtime.schedule import ProgramTaskList
from qoala.sim.csocket import ClassicalSocket
from qoala.sim.hostinterface import HostInterface
from qoala.sim.hostprocessor import HostProcessor
from qoala.sim.memory import ProgramMemory, SharedMemory, UnitModule
from qoala.sim.message import Message
from qoala.sim.process import IqoalaProcess
from qoala.util.tests import yield_from

MOCK_MESSAGE = Message(content=42)
MOCK_QNOS_RET_REG = "R0"
MOCK_QNOS_RET_VALUE = 7


@dataclass(eq=True, frozen=True)
class InterfaceEvent:
    peer: str
    msg: Message


class MockHostInterface(HostInterface):
    def __init__(self, shared_mem: Optional[SharedMemory] = None) -> None:
        self.send_events: List[InterfaceEvent] = []
        self.recv_events: List[InterfaceEvent] = []

        self.shared_mem = shared_mem

    def send_peer_msg(self, peer: str, msg: Message) -> None:
        self.send_events.append(InterfaceEvent(peer, msg))

    def receive_peer_msg(self, peer: str) -> Generator[EventExpression, None, Message]:
        self.recv_events.append(InterfaceEvent(peer, MOCK_MESSAGE))
        return MOCK_MESSAGE
        yield  # to make it behave as a generator

    def send_qnos_msg(self, msg: Message) -> None:
        self.send_events.append(InterfaceEvent("qnos", msg))

    def receive_qnos_msg(self) -> Generator[EventExpression, None, Message]:
        self.recv_events.append(InterfaceEvent("qnos", MOCK_MESSAGE))
        if self.shared_mem is not None:
            self.shared_mem.set_reg_value(MOCK_QNOS_RET_REG, MOCK_QNOS_RET_VALUE)
        return MOCK_MESSAGE
        yield  # to make it behave as a generator

    @property
    def name(self) -> str:
        return "mock"


def create_program(
    instrs: Optional[List[ClassicalIqoalaOp]] = None,
    subroutines: Optional[Dict[str, IqoalaSubroutine]] = None,
    meta: Optional[ProgramMeta] = None,
) -> IqoalaProgram:
    if instrs is None:
        instrs = []
    if subroutines is None:
        subroutines = {}
    if meta is None:
        meta = ProgramMeta.empty("prog")
    return IqoalaProgram(instructions=instrs, subroutines=subroutines, meta=meta)


def create_process(
    program: IqoalaProgram,
    interface: HostInterface,
    inputs: Optional[Dict[str, Any]] = None,
) -> IqoalaProcess:
    if inputs is None:
        inputs = {}
    prog_input = ProgramInput(values=inputs)
    instance = ProgramInstance(
        pid=0,
        program=program,
        inputs=prog_input,
        tasks=ProgramTaskList.empty(program),
    )
    mem = ProgramMemory(pid=0, unit_module=UnitModule.default_generic(2))
    process = IqoalaProcess(
        prog_instance=instance,
        prog_memory=mem,
        csockets={
            id: ClassicalSocket(interface, name)
            for (id, name) in program.meta.csockets.items()
        },
        epr_sockets=program.meta.epr_sockets,
        subroutines=program.subroutines,
        requests={},
        result=ProgramResult(values={}),
    )
    return process


def test_initialize():
    interface = MockHostInterface()
    processor = HostProcessor(interface)
    program = create_program()
    process = create_process(
        program, interface, inputs={"x": 1, "theta": 3.14, "name": "alice"}
    )

    processor.initialize(process)
    host_mem = process.prog_memory.host_mem
    assert host_mem.read("x") == 1
    assert host_mem.read("theta") == 3.14
    assert host_mem.read("name") == "alice"


def test_assign_cvalue():
    interface = MockHostInterface()
    processor = HostProcessor(interface)
    program = create_program(instrs=[AssignCValueOp("x", 3)])
    process = create_process(program, interface)
    processor.initialize(process)

    yield_from(processor.assign(process, 0))
    assert process.prog_memory.host_mem.read("x") == 3


def test_send_msg():
    interface = MockHostInterface()
    processor = HostProcessor(interface)
    meta = ProgramMeta.empty("alice")
    meta.csockets = {0: "bob"}
    program = create_program(instrs=[SendCMsgOp("bob", "msg")], meta=meta)
    process = create_process(program, interface, inputs={"bob": 0, "msg": 12})
    processor.initialize(process)

    yield_from(processor.assign(process, 0))
    assert interface.send_events[0] == InterfaceEvent("bob", Message(content=12))


def test_recv_msg():
    interface = MockHostInterface()
    processor = HostProcessor(interface)
    meta = ProgramMeta.empty("alice")
    meta.csockets = {0: "bob"}
    program = create_program(instrs=[ReceiveCMsgOp("bob", "msg")], meta=meta)
    process = create_process(program, interface, inputs={"bob": 0})
    processor.initialize(process)

    yield_from(processor.assign(process, 0))
    assert interface.recv_events[0] == InterfaceEvent("bob", MOCK_MESSAGE)
    assert process.prog_memory.host_mem.read("msg") == MOCK_MESSAGE.content


def test_add_cvalue():
    interface = MockHostInterface()
    processor = HostProcessor(interface)
    program = create_program(
        instrs=[
            AssignCValueOp("a", 2),
            AssignCValueOp("b", 3),
            AddCValueOp("sum", "a", "b"),
        ]
    )
    process = create_process(program, interface)
    processor.initialize(process)

    for i in range(len(program.instructions)):
        yield_from(processor.assign(process, i))

    assert process.prog_memory.host_mem.read("sum") == 5


def test_multiply_const():
    interface = MockHostInterface()
    processor = HostProcessor(interface)
    program = create_program(
        instrs=[AssignCValueOp("a", 4), MultiplyConstantCValueOp("result", "a", -1)]
    )
    process = create_process(program, interface)
    processor.initialize(process)

    for i in range(len(program.instructions)):
        yield_from(processor.assign(process, i))

    assert process.prog_memory.host_mem.read("result") == -4


def test_bit_cond_mult():
    interface = MockHostInterface()
    processor = HostProcessor(interface)
    program = create_program(
        instrs=[
            AssignCValueOp("var1", 4),
            AssignCValueOp("var2", 7),
            AssignCValueOp("cond1", 0),
            AssignCValueOp("cond2", 1),
            BitConditionalMultiplyConstantCValueOp("result1", "var1", "cond1", -1),
            BitConditionalMultiplyConstantCValueOp("result2", "var2", "cond2", -1),
        ]
    )
    process = create_process(program, interface)
    processor.initialize(process)

    for i in range(len(program.instructions)):
        yield_from(processor.assign(process, i))

    assert process.prog_memory.host_mem.read("result1") == 4
    assert process.prog_memory.host_mem.read("result2") == -7


def test_run_subroutine():
    interface = MockHostInterface()
    processor = HostProcessor(interface)

    subrt = Subroutine()
    iqoala_subrt = IqoalaSubroutine("subrt1", subrt, return_map={})

    program = create_program(
        instrs=[RunSubroutineOp(None, IqoalaVector([]), "subrt1")],
        subroutines={"subrt1": iqoala_subrt},
    )
    process = create_process(program, interface)
    processor.initialize(process)

    for i in range(len(program.instructions)):
        yield_from(processor.assign(process, i))

    # Non-async host processor should not have done any communciation.
    assert len(interface.send_events) == 0
    assert len(interface.recv_events) == 0


def test_run_subroutine_async():
    interface = MockHostInterface()
    processor = HostProcessor(interface, asynchronous=True)

    subrt = Subroutine()
    iqoala_subrt = IqoalaSubroutine("subrt1", subrt, return_map={})

    program = create_program(
        instrs=[RunSubroutineOp(None, IqoalaVector([]), "subrt1")],
        subroutines={"subrt1": iqoala_subrt},
    )
    process = create_process(program, interface)
    processor.initialize(process)

    for i in range(len(program.instructions)):
        yield_from(processor.assign(process, i))

    # Async host processor should have communicated with the qnos processor.
    assert interface.send_events[0] == InterfaceEvent("qnos", Message("subrt1"))
    assert interface.recv_events[0] == InterfaceEvent("qnos", MOCK_MESSAGE)


def test_run_subroutine_async_2():
    interface = MockHostInterface()
    processor = HostProcessor(interface, asynchronous=True)

    subrt_text = """
    set R0 {my_value}
    ret_reg R0
    """
    subrt = parse_text_subroutine(subrt_text)
    iqoala_subrt = IqoalaSubroutine(
        "subrt1", subrt, return_map={"m": IqoalaSharedMemLoc("R0")}
    )

    program = create_program(
        instrs=[
            AssignCValueOp("my_value", 16),
            RunSubroutineOp(None, IqoalaVector(["my_value"]), "subrt1"),
        ],
        subroutines={"subrt1": iqoala_subrt},
    )
    process = create_process(program, interface)

    # Make sure interface can mimick writing subroutine results to shared memory
    interface.shared_mem = process.prog_memory.shared_mem

    processor.initialize(process)

    for i in range(len(program.instructions)):
        yield_from(processor.assign(process, i))

    assert interface.send_events[0] == InterfaceEvent("qnos", Message("subrt1"))
    assert interface.recv_events[0] == InterfaceEvent("qnos", MOCK_MESSAGE)
    assert process.prog_memory.shared_mem.get_reg_value("R0") == MOCK_QNOS_RET_VALUE
    assert process.prog_memory.host_mem.read("m") == MOCK_QNOS_RET_VALUE


def test_return_result():
    interface = MockHostInterface()
    processor = HostProcessor(interface)
    program = create_program(
        instrs=[AssignCValueOp("result", 2), ReturnResultOp("result")]
    )
    process = create_process(program, interface)
    processor.initialize(process)

    for i in range(len(program.instructions)):
        yield_from(processor.assign(process, i))

    assert process.prog_memory.host_mem.read("result") == 2
    assert process.result.values == {"result": 2}


if __name__ == "__main__":
    test_initialize()
    test_assign_cvalue()
    test_send_msg()
    test_recv_msg()
    test_add_cvalue()
    test_multiply_const()
    test_bit_cond_mult()
    test_run_subroutine()
    test_run_subroutine_async()
    test_run_subroutine_async_2()
    test_return_result()



import netsquid as ns
import pytest
from netsquid.components import instructions as ns_instr
from netsquid.components.qprocessor import MissingInstructionError
from netsquid.nodes import Node
from netsquid.qubits import ketstates

from qoala.runtime.config import GenericQDeviceConfig, NVQDeviceConfig
from qoala.sim.build import build_generic_qprocessor, build_nv_qprocessor
from qoala.sim.constants import PI, PI_OVER_2
from qoala.sim.qdevice import (
    GenericPhysicalQuantumMemory,
    NonInitializedQubitError,
    NVPhysicalQuantumMemory,
    QDevice,
    QDeviceCommand,
    QDeviceType,
    UnsupportedQDeviceCommandError,
)
from qoala.util.tests import has_state, netsquid_run


def perfect_generic_qdevice(num_qubits: int) -> QDevice:
    cfg = GenericQDeviceConfig.perfect_config(num_qubits=num_qubits)
    processor = build_generic_qprocessor(name="processor", cfg=cfg)
    node = Node(name="alice", qmemory=processor)
    return QDevice(
        node=node,
        typ=QDeviceType.GENERIC,
        memory=GenericPhysicalQuantumMemory(num_qubits),
    )


def perfect_nv_qdevice(num_qubits: int) -> QDevice:
    cfg = NVQDeviceConfig.perfect_config(num_qubits=num_qubits)
    processor = build_nv_qprocessor(name="processor", cfg=cfg)
    node = Node(name="alice", qmemory=processor)
    return QDevice(
        node=node,
        typ=QDeviceType.NV,
        memory=NVPhysicalQuantumMemory(num_qubits),
    )


def test_static_generic():
    num_qubits = 3
    qdevice = perfect_generic_qdevice(num_qubits)

    assert qdevice.qprocessor.num_positions == num_qubits

    assert qdevice.typ == QDeviceType.GENERIC
    assert qdevice.qubit_count == num_qubits
    assert qdevice.comm_qubit_count == num_qubits
    assert qdevice.comm_qubit_ids == {i for i in range(num_qubits)}
    assert qdevice.mem_qubit_ids == {i for i in range(num_qubits)}
    assert qdevice.all_qubit_ids == {i for i in range(num_qubits)}


def test_static_nv():
    num_qubits = 3
    qdevice = perfect_nv_qdevice(num_qubits)

    assert qdevice.qprocessor.num_positions == num_qubits
    with pytest.raises(MissingInstructionError):
        qdevice.qprocessor.get_instruction_duration(ns_instr.INSTR_CNOT, [0, 1])
        qdevice.qprocessor.get_instruction_duration(ns_instr.INSTR_CXDIR, [1, 0])

    # Should not raise error:
    qdevice.qprocessor.get_instruction_duration(ns_instr.INSTR_CXDIR, [0, 1])

    assert qdevice.typ == QDeviceType.NV
    assert qdevice.qubit_count == num_qubits
    assert qdevice.comm_qubit_count == 1
    assert qdevice.comm_qubit_ids == {0}
    assert qdevice.mem_qubit_ids == {i for i in range(1, num_qubits)}
    assert qdevice.all_qubit_ids == {i for i in range(num_qubits)}


def test_initalize_generic():
    num_qubits = 3
    qdevice = perfect_generic_qdevice(num_qubits)

    # All qubits are not initalized yet.
    assert qdevice.get_local_qubit(0) is None
    assert qdevice.get_local_qubit(1) is None
    assert qdevice.get_local_qubit(2) is None

    # Initialize qubit 0.
    commands = [QDeviceCommand(ns_instr.INSTR_INIT, [0])]
    netsquid_run(qdevice.execute_commands(commands))
    ns.sim_run()

    # Qubit 0 should be initalized and have state |0>
    q0 = qdevice.get_local_qubit(0)
    assert has_state(q0, ketstates.s0)

    commands = [QDeviceCommand(ns_instr.INSTR_X, [0])]
    netsquid_run(qdevice.execute_commands(commands))
    ns.sim_run()

    q0 = qdevice.get_local_qubit(0)
    assert has_state(q0, ketstates.s1)

    # Qubits 1 and 2 are still not initalized.
    assert qdevice.get_local_qubit(1) is None
    assert qdevice.get_local_qubit(2) is None

    # Initialize qubit 1.
    commands = [QDeviceCommand(ns_instr.INSTR_INIT, [1])]
    netsquid_run(qdevice.execute_commands(commands))
    ns.sim_run()

    # Qubit 0 should still be in |1>.
    # Qubit 1 should be in |0>.
    q0 = qdevice.get_local_qubit(0)
    q1 = qdevice.get_local_qubit(1)
    assert has_state(q0, ketstates.s1)
    assert has_state(q1, ketstates.s0)

    # Test getting multiple qubits at the same time.
    [q0, q1, q2] = qdevice.get_local_qubits([0, 1, 2])
    assert has_state(q0, ketstates.s1)
    assert has_state(q1, ketstates.s0)
    assert q2 is None


def test_initalize_nv():
    num_qubits = 3
    qdevice = perfect_nv_qdevice(num_qubits)

    # All qubits are not initalized yet.
    assert qdevice.get_local_qubit(0) is None
    assert qdevice.get_local_qubit(1) is None
    assert qdevice.get_local_qubit(2) is None

    # Initialize qubit 0.
    commands = [QDeviceCommand(ns_instr.INSTR_INIT, [0])]
    netsquid_run(qdevice.execute_commands(commands))
    ns.sim_run()

    # Qubit 0 should be initalized and have state |0>
    q0 = qdevice.get_local_qubit(0)
    assert has_state(q0, ketstates.s0)

    with pytest.raises(UnsupportedQDeviceCommandError):
        commands = [QDeviceCommand(ns_instr.INSTR_H, [0])]
        netsquid_run(qdevice.execute_commands(commands))
        ns.sim_run()

    with pytest.raises(UnsupportedQDeviceCommandError):
        commands = [QDeviceCommand(ns_instr.INSTR_X, [0])]
        netsquid_run(qdevice.execute_commands(commands))
        ns.sim_run()

    commands = [QDeviceCommand(ns_instr.INSTR_ROT_X, [0], angle=PI)]
    netsquid_run(qdevice.execute_commands(commands))
    ns.sim_run()

    q0 = qdevice.get_local_qubit(0)
    assert has_state(q0, ketstates.s1)

    # Qubits 1 and 2 are still not initalized.
    assert qdevice.get_local_qubit(1) is None
    assert qdevice.get_local_qubit(2) is None

    # Initialize qubit 1.
    commands = [QDeviceCommand(ns_instr.INSTR_INIT, [1])]
    netsquid_run(qdevice.execute_commands(commands))
    ns.sim_run()

    # Qubit 0 should still be in |1>.
    # Qubit 1 should be in |0>.
    q0 = qdevice.get_local_qubit(0)
    q1 = qdevice.get_local_qubit(1)
    assert has_state(q0, ketstates.s1)
    assert has_state(q1, ketstates.s0)

    # Test getting multiple qubits at the same time.
    [q0, q1, q2] = qdevice.get_local_qubits([0, 1, 2])
    assert has_state(q0, ketstates.s1)
    assert has_state(q1, ketstates.s0)
    assert q2 is None


def test_rotations_generic():
    num_qubits = 3
    qdevice = perfect_generic_qdevice(num_qubits)

    # All qubits are not initalized yet.
    assert qdevice.get_local_qubit(0) is None
    assert qdevice.get_local_qubit(1) is None
    assert qdevice.get_local_qubit(2) is None

    # Initialize qubit 0 and do Y-rotation of PI.
    commands = [
        QDeviceCommand(ns_instr.INSTR_INIT, [0]),
        QDeviceCommand(ns_instr.INSTR_Y, [0]),
    ]
    netsquid_run(qdevice.execute_commands(commands))
    ns.sim_run()

    q0 = qdevice.get_local_qubit(0)
    assert has_state(q0, ketstates.s1)

    # Initialize qubit 1 and do H gate.
    commands = [
        QDeviceCommand(ns_instr.INSTR_INIT, [1]),
        QDeviceCommand(ns_instr.INSTR_H, [1]),
    ]
    netsquid_run(qdevice.execute_commands(commands))
    ns.sim_run()

    q1 = qdevice.get_local_qubit(1)
    assert has_state(q1, ketstates.h0)

    # Initialize qubit 2, do a H gate, and a Z-rotation of PI/2.
    commands = [
        QDeviceCommand(ns_instr.INSTR_INIT, [2]),
        QDeviceCommand(ns_instr.INSTR_H, [2]),
        QDeviceCommand(ns_instr.INSTR_ROT_Z, [2], angle=PI_OVER_2),
    ]
    netsquid_run(qdevice.execute_commands(commands))
    ns.sim_run()

    q2 = qdevice.get_local_qubit(2)
    assert has_state(q2, ketstates.y0)


def test_rotations_nv():
    num_qubits = 3
    qdevice = perfect_nv_qdevice(num_qubits)

    # All qubits are not initalized yet.
    assert qdevice.get_local_qubit(0) is None
    assert qdevice.get_local_qubit(1) is None
    assert qdevice.get_local_qubit(2) is None

    # NV QDevice does not support X-gate.
    with pytest.raises(UnsupportedQDeviceCommandError):
        commands = [
            QDeviceCommand(ns_instr.INSTR_INIT, [0]),
            QDeviceCommand(ns_instr.INSTR_X, [0]),
        ]
        netsquid_run(qdevice.execute_commands(commands))
        ns.sim_run()

    # Initialize qubit 0 and do X-rotation of PI.
    commands = [
        QDeviceCommand(ns_instr.INSTR_INIT, [0]),
        QDeviceCommand(ns_instr.INSTR_ROT_X, [0], angle=PI),
    ]
    netsquid_run(qdevice.execute_commands(commands))
    ns.sim_run()

    q0 = qdevice.get_local_qubit(0)
    assert has_state(q0, ketstates.s1)

    # Initialize qubit 1 and do a Y-rotation of PI/2.
    commands = [
        QDeviceCommand(ns_instr.INSTR_INIT, [1]),
        QDeviceCommand(ns_instr.INSTR_ROT_Y, [1], angle=PI_OVER_2),
    ]
    netsquid_run(qdevice.execute_commands(commands))
    ns.sim_run()

    q1 = qdevice.get_local_qubit(1)
    assert has_state(q1, ketstates.h0)

    # Initialize qubit 2, do a Y-rotation of PI/2, and a Z-rotation of PI/2.
    commands = [
        QDeviceCommand(ns_instr.INSTR_INIT, [2]),
        QDeviceCommand(ns_instr.INSTR_ROT_Y, [2], angle=PI_OVER_2),
        QDeviceCommand(ns_instr.INSTR_ROT_Z, [2], angle=PI_OVER_2),
    ]
    netsquid_run(qdevice.execute_commands(commands))
    ns.sim_run()

    q2 = qdevice.get_local_qubit(2)
    assert has_state(q2, ketstates.y0)


def test_measure_generic():
    num_qubits = 3
    qdevice = perfect_generic_qdevice(num_qubits)

    # Initialize qubit 0 and measure it.
    commands = [
        QDeviceCommand(ns_instr.INSTR_INIT, [0]),
        QDeviceCommand(ns_instr.INSTR_MEASURE, [0]),
    ]

    meas_outcome = netsquid_run(qdevice.execute_commands(commands))

    q0 = qdevice.get_local_qubit(0)
    # Applying the NetSquid measurement instruction should not discard the qubit.
    # (A QnosProcessor should do this manually!)
    assert q0 is not None
    assert has_state(q0, ketstates.s0)

    assert meas_outcome is not None
    assert meas_outcome == 0

    # Initialize qubit 1, apply X gate, and measure it.
    commands = [
        QDeviceCommand(ns_instr.INSTR_INIT, [1]),
        QDeviceCommand(ns_instr.INSTR_X, [1]),
        QDeviceCommand(ns_instr.INSTR_MEASURE, [1]),
    ]

    meas_outcome = netsquid_run(qdevice.execute_commands(commands))

    q1 = qdevice.get_local_qubit(1)
    # Applying the NetSquid measurement instruction should not discard the qubit.
    # (A QnosProcessor should do this manually!)
    assert q1 is not None
    assert has_state(q1, ketstates.s1)

    assert meas_outcome is not None
    assert meas_outcome == 1

    # Measure qubit 0 again.
    commands = [
        QDeviceCommand(ns_instr.INSTR_MEASURE, [0]),
    ]

    meas_outcome = netsquid_run(qdevice.execute_commands(commands))

    q0 = qdevice.get_local_qubit(0)
    assert q0 is not None
    assert has_state(q0, ketstates.s0)

    assert meas_outcome is not None
    assert meas_outcome == 0


def test_measure_nv():
    num_qubits = 3
    qdevice = perfect_nv_qdevice(num_qubits)

    # Initialize qubit 0 and measure it.
    commands = [
        QDeviceCommand(ns_instr.INSTR_INIT, [0]),
        QDeviceCommand(ns_instr.INSTR_MEASURE, [0]),
    ]

    meas_outcome = netsquid_run(qdevice.execute_commands(commands))

    q0 = qdevice.get_local_qubit(0)
    # Applying the NetSquid measurement instruction should not discard the qubit.
    # (A QnosProcessor should do this manually!)
    assert q0 is not None
    assert has_state(q0, ketstates.s0)

    assert meas_outcome is not None
    assert meas_outcome == 0

    # Initialize qubit 1, and try to measure it. Should raise an error since
    # only qubit 0 can be measured.
    with pytest.raises(UnsupportedQDeviceCommandError):
        commands = [
            QDeviceCommand(ns_instr.INSTR_INIT, [1]),
            QDeviceCommand(ns_instr.INSTR_MEASURE, [1]),
        ]
        meas_outcome = netsquid_run(qdevice.execute_commands(commands))

    # Measure qubit 0 again.
    commands = [QDeviceCommand(ns_instr.INSTR_MEASURE, [0])]
    meas_outcome = netsquid_run(qdevice.execute_commands(commands))

    q0 = qdevice.get_local_qubit(0)
    assert q0 is not None
    assert has_state(q0, ketstates.s0)

    assert meas_outcome is not None
    assert meas_outcome == 0


def test_two_qubit_gates_generic():
    num_qubits = 3
    qdevice = perfect_generic_qdevice(num_qubits)

    # All qubits are not initalized yet.
    assert qdevice.get_local_qubit(0) is None
    assert qdevice.get_local_qubit(1) is None
    assert qdevice.get_local_qubit(2) is None

    # Initialize qubits 0 and 2 and apply a CNOT
    commands = [
        QDeviceCommand(ns_instr.INSTR_INIT, [0]),
        QDeviceCommand(ns_instr.INSTR_INIT, [2]),
        QDeviceCommand(ns_instr.INSTR_CNOT, [0, 2]),
    ]
    netsquid_run(qdevice.execute_commands(commands))
    ns.sim_run()

    [q0, q1, q2] = qdevice.get_local_qubits([0, 1, 2])
    assert has_state(q0, ketstates.s0)
    assert q1 is None
    assert has_state(q2, ketstates.s0)

    # Apply an X to qubit 2, and a do CNOT between 2 and 0.
    commands = [
        QDeviceCommand(ns_instr.INSTR_X, [2]),
        QDeviceCommand(ns_instr.INSTR_CNOT, [2, 0]),
    ]
    netsquid_run(qdevice.execute_commands(commands))
    ns.sim_run()

    [q0, q1, q2] = qdevice.get_local_qubits([0, 1, 2])
    assert has_state(q0, ketstates.s1)
    assert q1 is None
    assert has_state(q2, ketstates.s1)


def test_two_qubit_gates_nv():
    num_qubits = 3
    qdevice = perfect_nv_qdevice(num_qubits)

    # All qubits are not initalized yet.
    assert qdevice.get_local_qubit(0) is None
    assert qdevice.get_local_qubit(1) is None
    assert qdevice.get_local_qubit(2) is None

    # Initialize qubits 0 and 2 and apply a CXDIR
    commands = [
        QDeviceCommand(ns_instr.INSTR_INIT, [0]),
        QDeviceCommand(ns_instr.INSTR_INIT, [2]),
        QDeviceCommand(ns_instr.INSTR_CXDIR, [0, 2], angle=PI_OVER_2),
    ]
    netsquid_run(qdevice.execute_commands(commands))
    ns.sim_run()

    # q0 is |0> so q2 was rotated with pi/2 around X
    [q0, q1, q2] = qdevice.get_local_qubits([0, 1, 2])
    assert has_state(q0, ketstates.s0)
    assert q1 is None
    assert has_state(q2, ketstates.y1)

    # Apply an X to qubit 0, and a do CNOT between 0 and 2.
    commands = [
        QDeviceCommand(ns_instr.INSTR_ROT_X, [0], angle=PI),
        QDeviceCommand(ns_instr.INSTR_CXDIR, [0, 2], angle=PI_OVER_2),
    ]
    netsquid_run(qdevice.execute_commands(commands))
    ns.sim_run()

    # q0 is |1> so q2 was rotated with -pi/2 around X
    [q0, q1, q2] = qdevice.get_local_qubits([0, 1, 2])
    assert has_state(q0, ketstates.s1)
    assert q1 is None
    assert has_state(q2, ketstates.s0)


def test_unsupported_commands_generic():
    num_qubits = 3
    qdevice = perfect_generic_qdevice(num_qubits)

    with pytest.raises(UnsupportedQDeviceCommandError):
        commands = [QDeviceCommand(ns_instr.INSTR_T, [0])]
        netsquid_run(qdevice.execute_commands(commands))


def test_unsupported_commands_nv():
    num_qubits = 3
    qdevice = perfect_nv_qdevice(num_qubits)

    with pytest.raises(UnsupportedQDeviceCommandError):
        commands = [QDeviceCommand(ns_instr.INSTR_X, [0])]
        netsquid_run(qdevice.execute_commands(commands))

    with pytest.raises(UnsupportedQDeviceCommandError):
        commands = [QDeviceCommand(ns_instr.INSTR_Y, [0])]
        netsquid_run(qdevice.execute_commands(commands))

    with pytest.raises(UnsupportedQDeviceCommandError):
        commands = [QDeviceCommand(ns_instr.INSTR_Z, [0])]
        netsquid_run(qdevice.execute_commands(commands))

    with pytest.raises(UnsupportedQDeviceCommandError):
        commands = [QDeviceCommand(ns_instr.INSTR_H, [0])]
        netsquid_run(qdevice.execute_commands(commands))

    with pytest.raises(UnsupportedQDeviceCommandError):
        commands = [QDeviceCommand(ns_instr.INSTR_X, [1])]
        netsquid_run(qdevice.execute_commands(commands))

    with pytest.raises(UnsupportedQDeviceCommandError):
        commands = [QDeviceCommand(ns_instr.INSTR_MEASURE, [1])]
        netsquid_run(qdevice.execute_commands(commands))

    with pytest.raises(UnsupportedQDeviceCommandError):
        commands = [QDeviceCommand(ns_instr.INSTR_MEASURE, [2])]
        netsquid_run(qdevice.execute_commands(commands))

    with pytest.raises(UnsupportedQDeviceCommandError):
        commands = [QDeviceCommand(ns_instr.INSTR_CXDIR, [0, 0])]
        netsquid_run(qdevice.execute_commands(commands))

    with pytest.raises(UnsupportedQDeviceCommandError):
        commands = [QDeviceCommand(ns_instr.INSTR_CXDIR, [1, 0])]
        netsquid_run(qdevice.execute_commands(commands))

    with pytest.raises(UnsupportedQDeviceCommandError):
        commands = [QDeviceCommand(ns_instr.INSTR_CXDIR, [1, 2])]
        netsquid_run(qdevice.execute_commands(commands))


def test_non_initalized():
    num_qubits = 3
    qdevice = perfect_generic_qdevice(num_qubits)

    with pytest.raises(NonInitializedQubitError):
        commands = [QDeviceCommand(ns_instr.INSTR_X, [0])]
        netsquid_run(qdevice.execute_commands(commands))

    with pytest.raises(NonInitializedQubitError):
        commands = [QDeviceCommand(ns_instr.INSTR_INIT, [0])]
        commands = [QDeviceCommand(ns_instr.INSTR_CNOT, [0, 1])]
        netsquid_run(qdevice.execute_commands(commands))


if __name__ == "__main__":
    test_static_generic()
    test_static_nv()
    test_initalize_generic()
    test_initalize_nv()
    test_rotations_generic()
    test_rotations_nv()
    test_measure_generic()
    test_measure_nv()
    test_two_qubit_gates_generic()
    test_two_qubit_gates_nv()
    test_unsupported_commands_generic()
    test_unsupported_commands_nv()
    test_non_initalized()



import abc
from dataclasses import dataclass
from typing import Any, Dict, List

from qoala.lang.iqoala import IqoalaProgram
from qoala.runtime.schedule import ProgramTaskList


class ProgramContext(abc.ABC):
    pass


@dataclass
class ProgramInput:
    values: Dict[str, Any]


@dataclass
class ProgramResult:
    values: Dict[str, Any]


@dataclass
class BatchInfo:
    """Description of a batch of program instances that should be executed."""

    program: IqoalaProgram
    inputs: List[ProgramInput]  # dict of inputs for each iteration
    num_iterations: int
    deadline: float
    tasks: ProgramTaskList
    num_qubits: int  # TODO: replace this by unit module


@dataclass
class ProgramInstance:
    """A running program"""

    pid: int
    program: IqoalaProgram
    inputs: ProgramInput
    tasks: ProgramTaskList


@dataclass
class ProgramBatch:
    batch_id: int
    info: BatchInfo
    instances: List[ProgramInstance]


@dataclass
class BatchResult:
    batch_id: int
    results: List[ProgramResult]



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, Generator, List, Optional, Tuple, Type

import netsquid as ns
from netqasm.lang.instr.core import CreateEPRInstruction, RecvEPRInstruction
from netqasm.lang.parsing import parse_text_subroutine
from netqasm.sdk.build_epr import SER_RESPONSE_KEEP_LEN
from netsquid.components import QuantumProcessor
from netsquid.nodes import Node
from netsquid.qubits import ketstates
from netsquid.qubits.ketstates import BellIndex
from netsquid_magic.link_layer import (
    MagicLinkLayerProtocolWithSignaling,
    SingleClickTranslationUnit,
)
from netsquid_magic.magic_distributor import PerfectStateMagicDistributor
from qlink_interface import ReqCreateBase, ResCreateAndKeep
from qlink_interface.interface import ResCreate

from pydynaa import EventExpression
from qoala.lang.iqoala import (
    AssignCValueOp,
    ClassicalIqoalaOp,
    IqoalaParser,
    IqoalaProgram,
    IqoalaRequest,
    IqoalaSubroutine,
    IQoalaSubroutineParser,
    IqoalaVector,
    ProgramMeta,
    ReceiveCMsgOp,
    RunSubroutineOp,
    SendCMsgOp,
)
from qoala.runtime.config import GenericQDeviceConfig
from qoala.runtime.environment import (
    GlobalEnvironment,
    GlobalNodeInfo,
    LocalEnvironment,
)
from qoala.runtime.program import ProgramInput, ProgramInstance, ProgramResult
from qoala.runtime.schedule import ProgramTaskList
from qoala.sim.build import build_generic_qprocessor
from qoala.sim.csocket import ClassicalSocket
from qoala.sim.egp import EgpProtocol
from qoala.sim.hostinterface import HostInterface
from qoala.sim.memmgr import AllocError, MemoryManager
from qoala.sim.memory import ProgramMemory, SharedMemory, Topology, UnitModule
from qoala.sim.message import Message
from qoala.sim.netstackinterface import NetstackInterface
from qoala.sim.process import IqoalaProcess
from qoala.sim.procnode import ProcNode
from qoala.sim.qdevice import (
    PhysicalQuantumMemory,
    QDevice,
    QDeviceCommand,
    QDeviceType,
)
from qoala.sim.qnosinterface import QnosInterface
from qoala.sim.requests import (
    EprCreateType,
    NetstackCreateRequest,
    NetstackReceiveRequest,
)
from qoala.util.tests import has_multi_state, netsquid_run

MOCK_MESSAGE = Message(content=42)
MOCK_QNOS_RET_REG = "R0"
MOCK_QNOS_RET_VALUE = 7


@dataclass(eq=True, frozen=True)
class InterfaceEvent:
    peer: str
    msg: Message


@dataclass(eq=True, frozen=True)
class FlushEvent:
    pass


@dataclass(eq=True, frozen=True)
class SignalEvent:
    pass


class MockNetstackInterface(NetstackInterface):
    def __init__(
        self,
        local_env: LocalEnvironment,
        qdevice: QDevice,
        memmgr: MemoryManager,
        mock_result: ResCreate,
    ) -> None:
        self._qdevice = qdevice
        self._local_env = local_env
        self._memmgr = memmgr
        self._mock_result = mock_result

        self._requests_put: Dict[int, List[ReqCreateBase]] = {}
        self._awaited_result_ck: List[int] = []  # list of remote ids
        self._awaited_mem_free_sig_count: int = 0

    def put_request(self, remote_id: int, request: ReqCreateBase) -> None:
        if remote_id not in self._requests_put:
            self._requests_put[remote_id] = []
        self._requests_put[remote_id].append(request)

    def await_result_create_keep(
        self, remote_id: int
    ) -> Generator[EventExpression, None, ResCreateAndKeep]:
        self._awaited_result_ck.append(remote_id)
        return self._mock_result
        yield  # to make this behave as a generator

    def await_memory_freed_signal(
        self, pid: int, virt_id: int
    ) -> Generator[EventExpression, None, None]:
        raise AllocError
        yield  # to make this behave as a generator

    def send_qnos_msg(self, msg: Message) -> None:
        return None

    def send_peer_msg(self, peer: str, msg: Message) -> None:
        return None

    def receive_peer_msg(self, peer: str) -> Generator[EventExpression, None, Message]:
        return None
        yield  # to make this behave as a generator

    def reset(self) -> None:
        self._requests_put = {}
        self._awaited_result_ck = []
        self._awaited_mem_free_sig_count = 0


class MockQDevice(QDevice):
    def __init__(self, topology: Topology) -> None:
        self._memory = PhysicalQuantumMemory(topology.comm_ids, topology.mem_ids)

        self._executed_commands: List[QDeviceCommand] = []

    @property
    def typ(self) -> QDeviceType:
        return QDeviceType.GENERIC

    def set_mem_pos_in_use(self, id: int, in_use: bool) -> None:
        pass

    def execute_commands(
        self, commands: List[QDeviceCommand]
    ) -> Generator[EventExpression, None, Optional[int]]:
        self._executed_commands.extend(commands)
        return None
        yield

    def reset(self) -> None:
        self._executed_commands = []


@dataclass
class MockNetstackResultInfo:
    pid: int
    array_id: int
    start_idx: int
    end_idx: int


class MockQnosInterface(QnosInterface):
    def __init__(
        self,
        qdevice: QDevice,
        netstack_result_info: Optional[MockNetstackResultInfo] = None,
    ) -> None:
        self.send_events: List[InterfaceEvent] = []
        self.recv_events: List[InterfaceEvent] = []
        self.flush_events: List[FlushEvent] = []
        self.signal_events: List[SignalEvent] = []

        self._qdevice = qdevice
        self._memmgr = MemoryManager("alice", self._qdevice)

        self.netstack_result_info: Optional[
            MockNetstackResultInfo
        ] = netstack_result_info

    def send_peer_msg(self, peer: str, msg: Message) -> None:
        self.send_events.append(InterfaceEvent(peer, msg))

    def receive_peer_msg(self, peer: str) -> Generator[EventExpression, None, Message]:
        self.recv_events.append(InterfaceEvent(peer, MOCK_MESSAGE))
        return MOCK_MESSAGE
        yield  # to make it behave as a generator

    def send_host_msg(self, msg: Message) -> None:
        self.send_events.append(InterfaceEvent("host", msg))

    def receive_host_msg(self) -> Generator[EventExpression, None, Message]:
        self.recv_events.append(InterfaceEvent("host", MOCK_MESSAGE))
        return MOCK_MESSAGE
        yield  # to make it behave as a generator

    def send_netstack_msg(self, msg: Message) -> None:
        self.send_events.append(InterfaceEvent("netstack", msg))

    def receive_netstack_msg(self) -> Generator[EventExpression, None, Message]:
        self.recv_events.append(InterfaceEvent("netstack", MOCK_MESSAGE))
        if self.netstack_result_info is not None:
            mem = self.memmgr._processes[
                self.netstack_result_info.pid
            ].prog_memory.shared_mem
            array_id = self.netstack_result_info.array_id
            start_idx = self.netstack_result_info.start_idx
            end_idx = self.netstack_result_info.end_idx
            for i in range(start_idx, end_idx):
                mem.set_array_value(array_id, i, 42)
        return MOCK_MESSAGE
        yield  # to make it behave as a generator

    def flush_netstack_msgs(self) -> None:
        self.flush_events.append(FlushEvent())

    def signal_memory_freed(self) -> None:
        self.signal_events.append(SignalEvent())

    @property
    def name(self) -> str:
        return "mock"


class MockHostInterface(HostInterface):
    def __init__(self, shared_mem: Optional[SharedMemory] = None) -> None:
        self.send_events: List[InterfaceEvent] = []
        self.recv_events: List[InterfaceEvent] = []

        self.shared_mem = shared_mem

    def send_peer_msg(self, peer: str, msg: Message) -> None:
        self.send_events.append(InterfaceEvent(peer, msg))

    def receive_peer_msg(self, peer: str) -> Generator[EventExpression, None, Message]:
        self.recv_events.append(InterfaceEvent(peer, MOCK_MESSAGE))
        return MOCK_MESSAGE
        yield  # to make it behave as a generator

    def send_qnos_msg(self, msg: Message) -> None:
        self.send_events.append(InterfaceEvent("qnos", msg))

    def receive_qnos_msg(self) -> Generator[EventExpression, None, Message]:
        self.recv_events.append(InterfaceEvent("qnos", MOCK_MESSAGE))
        if self.shared_mem is not None:
            self.shared_mem.set_reg_value(MOCK_QNOS_RET_REG, MOCK_QNOS_RET_VALUE)
        return MOCK_MESSAGE
        yield  # to make it behave as a generator

    @property
    def name(self) -> str:
        return "mock"


def create_program(
    instrs: Optional[List[ClassicalIqoalaOp]] = None,
    subroutines: Optional[Dict[str, IqoalaSubroutine]] = None,
    requests: Optional[Dict[str, IqoalaRequest]] = None,
    meta: Optional[ProgramMeta] = None,
) -> IqoalaProgram:
    if instrs is None:
        instrs = []
    if subroutines is None:
        subroutines = {}

    if requests is None:
        requests = {}
    if meta is None:
        meta = ProgramMeta.empty("prog")
    return IqoalaProgram(
        instructions=instrs, subroutines=subroutines, meta=meta, requests=requests
    )


def create_process(
    pid: int,
    program: IqoalaProgram,
    unit_module: UnitModule,
    host_interface: HostInterface,
    inputs: Optional[Dict[str, Any]] = None,
) -> IqoalaProcess:
    if inputs is None:
        inputs = {}
    prog_input = ProgramInput(values=inputs)
    instance = ProgramInstance(
        pid=pid,
        program=program,
        inputs=prog_input,
        tasks=ProgramTaskList.empty(program),
    )
    mem = ProgramMemory(pid=0, unit_module=unit_module)

    process = IqoalaProcess(
        prog_instance=instance,
        prog_memory=mem,
        csockets={
            id: ClassicalSocket(host_interface, name)
            for (id, name) in program.meta.csockets.items()
        },
        epr_sockets=program.meta.epr_sockets,
        subroutines=program.subroutines,
        requests=program.requests,
        result=ProgramResult(values={}),
    )
    return process


def create_qprocessor(name: str, num_qubits: int) -> QuantumProcessor:
    cfg = GenericQDeviceConfig.perfect_config(num_qubits=num_qubits)
    return build_generic_qprocessor(name=f"{name}_processor", cfg=cfg)


def create_global_env(
    num_qubits: int, names: List[str] = ["alice", "bob", "charlie"]
) -> GlobalEnvironment:

    env = GlobalEnvironment()
    for i, name in enumerate(names):
        env.add_node(i, GlobalNodeInfo(name, i))
    return env


def create_procnode(
    name: str,
    env: GlobalEnvironment,
    num_qubits: int,
    procnode_cls: Type[ProcNode] = ProcNode,
    asynchronous: bool = False,
) -> ProcNode:
    alice_qprocessor = create_qprocessor(name, num_qubits)

    node_id = env.get_node_id(name)
    procnode = procnode_cls(
        name=name,
        global_env=env,
        qprocessor=alice_qprocessor,
        node_id=node_id,
        asynchronous=asynchronous,
    )

    return procnode


def simple_subroutine(name: str, subrt_text: str) -> IqoalaSubroutine:
    subrt = parse_text_subroutine(subrt_text)
    return IqoalaSubroutine(name, subrt, return_map={})


def parse_iqoala_subroutines(subrt_text: str) -> IqoalaSubroutine:
    return IQoalaSubroutineParser(subrt_text).parse()


def create_egp_protocols(node1: Node, node2: Node) -> Tuple[EgpProtocol, EgpProtocol]:
    link_dist = PerfectStateMagicDistributor(nodes=[node1, node2], state_delay=1000.0)
    link_prot = MagicLinkLayerProtocolWithSignaling(
        nodes=[node1, node2],
        magic_distributor=link_dist,
        translation_unit=SingleClickTranslationUnit(),
    )
    return EgpProtocol(node1, link_prot), EgpProtocol(node2, link_prot)


def test_initialize():
    num_qubits = 3
    topology = Topology(comm_ids={0, 1, 2}, mem_ids={0, 1, 2})

    global_env = create_global_env(num_qubits)
    local_env = LocalEnvironment(global_env, global_env.get_node_id("alice"))
    procnode = create_procnode("alice", global_env, num_qubits)
    procnode.qdevice = MockQDevice(topology)

    procnode.host.interface = MockHostInterface()

    mock_result = ResCreateAndKeep(bell_state=BellIndex.B01)
    procnode.netstack.interface = MockNetstackInterface(
        local_env, procnode.qdevice, procnode.memmgr, mock_result
    )

    host_processor = procnode.host.processor
    qnos_processor = procnode.qnos.processor
    netstack_processor = procnode.netstack.processor

    unit_module = UnitModule.default_generic(num_qubits)

    instrs = [AssignCValueOp("x", 3)]
    subrt1 = simple_subroutine(
        "subrt1",
        """
    set R5 42
    """,
    )

    program = create_program(instrs=instrs, subroutines={"subrt1": subrt1})
    process = create_process(
        pid=0,
        program=program,
        unit_module=unit_module,
        host_interface=procnode.host._interface,
        inputs={"x": 1, "theta": 3.14, "name": "alice"},
    )
    procnode.add_process(process)

    host_processor.initialize(process)
    host_mem = process.prog_memory.host_mem
    assert host_mem.read("x") == 1
    assert host_mem.read("theta") == 3.14
    assert host_mem.read("name") == "alice"

    request = NetstackCreateRequest(
        remote_id=1,
        epr_socket_id=0,
        typ=EprCreateType.CREATE_KEEP,
        num_pairs=1,
        fidelity=1.0,
        virt_qubit_ids=[0],
        result_array_addr=0,
    )

    netsquid_run(host_processor.assign(process, instr_idx=0))
    netsquid_run(qnos_processor.assign(process, "subrt1", 0))

    process.shared_mem.init_new_array(0, SER_RESPONSE_KEEP_LEN * 1)
    netsquid_run(netstack_processor.assign(process, request))

    assert process.host_mem.read("x") == 3
    assert process.shared_mem.get_reg_value("R5") == 42
    assert procnode.memmgr.phys_id_for(pid=process.pid, virt_id=0) == 0


def test_2():
    num_qubits = 3
    topology = Topology(comm_ids={0, 1, 2}, mem_ids={0, 1, 2})

    global_env = create_global_env(num_qubits)
    procnode = create_procnode("alice", global_env, num_qubits)
    procnode.qdevice = MockQDevice(topology)

    # procnode.host.interface = MockHostInterface()

    # mock_result = ResCreateAndKeep(bell_state=BellIndex.B01)
    # procnode.netstack.interface = MockNetstackInterface(
    #     local_env, procnode.qdevice, procnode.memmgr, mock_result
    # )

    host_processor = procnode.host.processor
    qnos_processor = procnode.qnos.processor

    unit_module = UnitModule.default_generic(num_qubits)

    instrs = [RunSubroutineOp(None, IqoalaVector([]), "subrt1")]
    subroutines = parse_iqoala_subroutines(
        """
SUBROUTINE subrt1
    params: 
    returns: R5 -> result
    request:
  NETQASM_START
    set R5 42
    ret_reg R5
  NETQASM_END
    """
    )

    program = create_program(instrs=instrs, subroutines=subroutines)
    process = create_process(
        pid=0,
        program=program,
        unit_module=unit_module,
        host_interface=procnode.host._interface,
        inputs={"x": 1, "theta": 3.14, "name": "alice"},
    )
    procnode.add_process(process)

    host_processor.initialize(process)

    netsquid_run(host_processor.assign(process, instr_idx=0))
    netsquid_run(qnos_processor.assign(process, "subrt1", 0))
    host_processor.copy_subroutine_results(process, "subrt1")

    assert process.host_mem.read("result") == 42
    assert process.shared_mem.get_reg_value("R5") == 42


def test_2_async():
    num_qubits = 3
    topology = Topology(comm_ids={0, 1, 2}, mem_ids={0, 1, 2})

    global_env = create_global_env(num_qubits)
    procnode = create_procnode("alice", global_env, num_qubits, asynchronous=True)
    procnode.qdevice = MockQDevice(topology)

    host_processor = procnode.host.processor
    qnos_processor = procnode.qnos.processor

    unit_module = UnitModule.default_generic(num_qubits)

    instrs = [RunSubroutineOp(None, IqoalaVector([]), "subrt1")]
    subroutines = parse_iqoala_subroutines(
        """
SUBROUTINE subrt1
    params: 
    returns: R5 -> result
    request:
  NETQASM_START
    set R5 42
    ret_reg R5
  NETQASM_END
    """
    )

    program = create_program(instrs=instrs, subroutines=subroutines)
    process = create_process(
        pid=0,
        program=program,
        unit_module=unit_module,
        host_interface=procnode.host._interface,
        inputs={"x": 1, "theta": 3.14, "name": "alice"},
    )
    procnode.add_process(process)

    host_processor.initialize(process)

    def host_run() -> Generator[EventExpression, None, None]:
        yield from host_processor.assign(process, instr_idx=0)

    def qnos_run() -> Generator[EventExpression, None, None]:
        yield from qnos_processor.assign(process, "subrt1", 0)
        # Mock sending signal back to Host that subroutine has finished.
        qnos_processor._interface.send_host_msg(Message(None))

    procnode.host.run = host_run
    procnode.qnos.run = qnos_run
    procnode.start()
    ns.sim_run()

    assert process.host_mem.read("result") == 42
    assert process.shared_mem.get_reg_value("R5") == 42


def test_classical_comm():
    num_qubits = 3

    global_env = create_global_env(num_qubits)

    class TestProcNode(ProcNode):
        def run(self) -> Generator[EventExpression, None, None]:
            process = self.memmgr.get_process(0)
            yield from self.host.processor.assign(process, 0)

    alice_procnode = create_procnode(
        "alice", global_env, num_qubits, procnode_cls=TestProcNode
    )
    bob_procnode = create_procnode(
        "bob", global_env, num_qubits, procnode_cls=TestProcNode
    )

    alice_host_processor = alice_procnode.host.processor
    bob_host_processor = bob_procnode.host.processor

    unit_module = UnitModule.default_generic(num_qubits)

    alice_instrs = [SendCMsgOp("csocket_id", "message")]
    alice_meta = ProgramMeta(
        name="alice",
        parameters=["csocket_id", "message"],
        csockets={0: "bob"},
        epr_sockets={},
    )
    alice_program = create_program(instrs=alice_instrs, meta=alice_meta)
    alice_process = create_process(
        pid=0,
        program=alice_program,
        unit_module=unit_module,
        host_interface=alice_procnode.host._interface,
        inputs={"csocket_id": 0, "message": 1337},
    )
    alice_procnode.add_process(alice_process)
    alice_host_processor.initialize(alice_process)

    bob_instrs = [ReceiveCMsgOp("csocket_id", "result")]
    bob_meta = ProgramMeta(
        name="bob", parameters=["csocket_id"], csockets={0: "alice"}, epr_sockets={}
    )
    bob_program = create_program(instrs=bob_instrs, meta=bob_meta)
    bob_process = create_process(
        pid=0,
        program=bob_program,
        unit_module=unit_module,
        host_interface=bob_procnode.host._interface,
        inputs={"csocket_id": 0},
    )
    bob_procnode.add_process(bob_process)
    bob_host_processor.initialize(bob_process)

    alice_procnode.connect_to(bob_procnode)

    # First start Bob, since Alice won't yield on anything (she only does a Send
    # instruction) and therefore calling 'start()' on alice completes her whole
    # protocol while Bob's interface has not even been started.
    bob_procnode.start()
    alice_procnode.start()
    ns.sim_run()

    assert bob_process.host_mem.read("result") == 1337


def test_classical_comm_three_nodes():
    num_qubits = 3

    global_env = create_global_env(num_qubits)

    class SenderProcNode(ProcNode):
        def run(self) -> Generator[EventExpression, None, None]:
            process = self.memmgr.get_process(0)
            yield from self.host.processor.assign(process, 0)

    class ReceiverProcNode(ProcNode):
        def run(self) -> Generator[EventExpression, None, None]:
            process = self.memmgr.get_process(0)
            yield from self.host.processor.assign(process, 0)
            yield from self.host.processor.assign(process, 1)

    alice_procnode = create_procnode(
        "alice", global_env, num_qubits, procnode_cls=SenderProcNode
    )
    bob_procnode = create_procnode(
        "bob", global_env, num_qubits, procnode_cls=SenderProcNode
    )
    charlie_procnode = create_procnode(
        "charlie", global_env, num_qubits, procnode_cls=ReceiverProcNode
    )

    alice_host_processor = alice_procnode.host.processor
    bob_host_processor = bob_procnode.host.processor
    charlie_host_processor = charlie_procnode.host.processor

    unit_module = UnitModule.default_generic(num_qubits)

    alice_instrs = [SendCMsgOp("csocket_id", "message")]
    alice_meta = ProgramMeta(
        name="alice",
        parameters=["csocket_id", "message"],
        csockets={0: "charlie"},
        epr_sockets={},
    )
    alice_program = create_program(instrs=alice_instrs, meta=alice_meta)
    alice_process = create_process(
        pid=0,
        program=alice_program,
        unit_module=unit_module,
        host_interface=alice_procnode.host._interface,
        inputs={"csocket_id": 0, "message": 1337},
    )
    alice_procnode.add_process(alice_process)
    alice_host_processor.initialize(alice_process)

    bob_instrs = [SendCMsgOp("csocket_id", "message")]
    bob_meta = ProgramMeta(
        name="bob",
        parameters=["csocket_id", "message"],
        csockets={0: "charlie"},
        epr_sockets={},
    )
    bob_program = create_program(instrs=bob_instrs, meta=bob_meta)
    bob_process = create_process(
        pid=0,
        program=bob_program,
        unit_module=unit_module,
        host_interface=bob_procnode.host._interface,
        inputs={"csocket_id": 0, "message": 42},
    )
    bob_procnode.add_process(bob_process)
    bob_host_processor.initialize(bob_process)

    charlie_instrs = [
        ReceiveCMsgOp("csocket_id_alice", "result_alice"),
        ReceiveCMsgOp("csocket_id_bob", "result_bob"),
    ]
    charlie_meta = ProgramMeta(
        name="bob",
        parameters=["csocket_id_alice", "csocket_id_bob"],
        csockets={0: "alice", 1: "bob"},
        epr_sockets={},
    )
    charlie_program = create_program(instrs=charlie_instrs, meta=charlie_meta)
    charlie_process = create_process(
        pid=0,
        program=charlie_program,
        unit_module=unit_module,
        host_interface=charlie_procnode.host._interface,
        inputs={"csocket_id_alice": 0, "csocket_id_bob": 1},
    )
    charlie_procnode.add_process(charlie_process)
    charlie_host_processor.initialize(charlie_process)

    alice_procnode.connect_to(charlie_procnode)
    bob_procnode.connect_to(charlie_procnode)

    # First start Charlie, since Alice and Bob don't yield on anything.
    charlie_procnode.start()
    alice_procnode.start()
    bob_procnode.start()
    ns.sim_run()

    assert charlie_process.host_mem.read("result_alice") == 1337
    assert charlie_process.host_mem.read("result_bob") == 42


def test_epr():
    num_qubits = 3

    global_env = create_global_env(num_qubits)
    alice_id = global_env.get_node_id("alice")
    bob_id = global_env.get_node_id("bob")

    class TestProcNode(ProcNode):
        def run(self) -> Generator[EventExpression, None, None]:
            process = self.memmgr.get_process(0)
            request = process.requests["req"]
            yield from self.netstack.processor.assign(process, request)

    alice_procnode = create_procnode(
        "alice", global_env, num_qubits, procnode_cls=TestProcNode
    )
    bob_procnode = create_procnode(
        "bob", global_env, num_qubits, procnode_cls=TestProcNode
    )

    alice_host_processor = alice_procnode.host.processor
    bob_host_processor = bob_procnode.host.processor

    unit_module = UnitModule.default_generic(num_qubits)

    alice_instrs = [SendCMsgOp("csocket_id", "message")]
    alice_meta = ProgramMeta(
        name="alice",
        parameters=["csocket_id", "message"],
        csockets={0: "bob"},
        epr_sockets={},
    )
    alice_program = create_program(instrs=alice_instrs, meta=alice_meta)
    alice_process = create_process(
        pid=0,
        program=alice_program,
        unit_module=unit_module,
        host_interface=alice_procnode.host._interface,
        inputs={"csocket_id": 0, "message": 1337},
    )
    alice_procnode.add_process(alice_process)
    alice_host_processor.initialize(alice_process)

    bob_instrs = [ReceiveCMsgOp("csocket_id", "result")]
    bob_meta = ProgramMeta(
        name="bob", parameters=["csocket_id"], csockets={0: "alice"}, epr_sockets={}
    )
    bob_program = create_program(instrs=bob_instrs, meta=bob_meta)
    bob_process = create_process(
        pid=0,
        program=bob_program,
        unit_module=unit_module,
        host_interface=bob_procnode.host._interface,
        inputs={"csocket_id": 0},
    )
    bob_procnode.add_process(bob_process)
    bob_host_processor.initialize(bob_process)

    alice_procnode.connect_to(bob_procnode)

    alice_request = NetstackCreateRequest(
        remote_id=bob_id,
        epr_socket_id=0,
        typ=EprCreateType.CREATE_KEEP,
        num_pairs=1,
        fidelity=1.0,
        virt_qubit_ids=[0],
        result_array_addr=0,
    )

    bob_request = NetstackReceiveRequest(
        remote_id=alice_id,
        epr_socket_id=0,
        typ=EprCreateType.CREATE_KEEP,
        num_pairs=1,
        fidelity=1.0,
        virt_qubit_ids=[0],
        result_array_addr=0,
    )

    alice_egp, bob_egp = create_egp_protocols(alice_procnode.node, bob_procnode.node)
    alice_procnode.egpmgr.add_egp(bob_id, alice_egp)
    bob_procnode.egpmgr.add_egp(alice_id, bob_egp)
    alice_process.requests = {"req": alice_request}
    bob_process.requests = {"req": bob_request}

    alice_process.shared_mem.init_new_array(0, 10)
    bob_process.shared_mem.init_new_array(0, 10)

    # First start Bob, since Alice won't yield on anything (she only does a Send
    # instruction) and therefore calling 'start()' on alice completes her whole
    # protocol while Bob's interface has not even been started.
    bob_procnode.start()
    alice_procnode.start()
    ns.sim_run()

    assert alice_procnode.memmgr.phys_id_for(pid=0, virt_id=0) == 0
    assert bob_procnode.memmgr.phys_id_for(pid=0, virt_id=0) == 0

    alice_qubit = alice_procnode.qdevice.get_local_qubit(0)
    bob_qubit = bob_procnode.qdevice.get_local_qubit(0)
    assert has_multi_state([alice_qubit, bob_qubit], ketstates.b00)


def test_whole_program():
    server_text = """
META_START
    name: server
    parameters: client_id
    csockets: 0 -> client
    epr_sockets: 0 -> client
META_END

remote_id = assign_cval() : {client_id}
run_subroutine(vec<remote_id>) : subrt1

SUBROUTINE subrt1
    params: remote_id
    returns:
    request: req1
  NETQASM_START
    array 10 @0
    recv_epr C15 C0 C1 C0
    wait_all @0[0:10]
  NETQASM_END

REQUEST req1
  role: receive
  remote_id: {client_id}
  epr_socket_id: 0
  typ: create_keep
  num_pairs: 1
  fidelity: 1.0
  virt_qubit_ids: 0
  result_array_addr: 0
    """
    server_program = IqoalaParser(server_text).parse()

    num_qubits = 3
    unit_module = UnitModule.default_generic(num_qubits)
    global_env = create_global_env(num_qubits, names=["client", "server"])
    server_id = global_env.get_node_id("server")
    client_id = global_env.get_node_id("client")

    class TestProcNode(ProcNode):
        def run(self) -> Generator[EventExpression, None, None]:
            process = self.memmgr.get_process(0)
            self.scheduler.initialize_process(process)

            subrt1 = process.subroutines["subrt1"]
            epr_instr_idx = None
            for i, instr in enumerate(subrt1.subroutine.instructions):
                if isinstance(instr, CreateEPRInstruction) or isinstance(
                    instr, RecvEPRInstruction
                ):
                    epr_instr_idx = i
                    break

            for i in range(epr_instr_idx):
                yield from self.qnos.processor.assign(process, "subrt1", i)

            request = process.requests["req1"].request
            print("hello 1?")
            yield from self.netstack.processor.assign(process, request)

            # wait instr
            yield from self.qnos.processor.assign(process, "subrt1", epr_instr_idx + 1)
            print("hello 2?")
            # yield from self.netstack.processor.assign(process, request)

    server_procnode = create_procnode(
        "server", global_env, num_qubits, procnode_cls=TestProcNode
    )
    server_process = create_process(
        pid=0,
        program=server_program,
        unit_module=unit_module,
        host_interface=server_procnode.host._interface,
        inputs={"client_id": client_id, "csocket_id": 0, "message": 1337},
    )
    server_procnode.add_process(server_process)

    client_text = """
META_START
    name: client
    parameters: server_id
    csockets: 0 -> server
    epr_sockets: 0 -> server
META_END

remote_id = assign_cval() : {server_id}
run_subroutine(vec<remote_id>) : subrt1

SUBROUTINE subrt1
    params: remote_id
    returns:
    request: req1
  NETQASM_START
    set C10 10
    array C10 @0
    create_epr C15 C0 C1 C2 C0
    wait_all @0[0:10]
  NETQASM_END

REQUEST req1
  role: create
  remote_id: {server_id}
  epr_socket_id: 0
  typ: create_keep
  num_pairs: 1
  fidelity: 1.0
  virt_qubit_ids: 0
  result_array_addr: 0
    """
    client_program = IqoalaParser(client_text).parse()
    client_procnode = create_procnode(
        "client", global_env, num_qubits, procnode_cls=TestProcNode
    )
    client_process = create_process(
        pid=0,
        program=client_program,
        unit_module=unit_module,
        host_interface=client_procnode.host._interface,
        inputs={"server_id": server_id, "csocket_id": 0, "message": 1337},
    )
    client_procnode.add_process(client_process)

    client_egp, server_egp = create_egp_protocols(
        client_procnode.node, server_procnode.node
    )
    client_procnode.egpmgr.add_egp(server_id, client_egp)
    server_procnode.egpmgr.add_egp(client_id, server_egp)

    client_procnode.connect_to(server_procnode)

    # client_egp._ll_prot.start()
    server_procnode.start()
    client_procnode.start()
    ns.sim_run()

    assert client_procnode.memmgr.phys_id_for(pid=0, virt_id=0) == 0
    assert server_procnode.memmgr.phys_id_for(pid=0, virt_id=0) == 0
    client_qubit = client_procnode.qdevice.get_local_qubit(0)
    server_qubit = server_procnode.qdevice.get_local_qubit(0)
    assert has_multi_state([client_qubit, server_qubit], ketstates.b00)

    # assert client_procnode.memmgr.phys_id_for(pid=0, virt_id=1) == 1
    # assert server_procnode.memmgr.phys_id_for(pid=0, virt_id=1) == 1


if __name__ == "__main__":
    test_initialize()
    test_2()
    test_2_async()
    test_classical_comm()
    test_classical_comm_three_nodes()
    test_epr()
    test_whole_program()



from __future__ import annotations

from typing import Any, List

import yaml
from pydantic import BaseModel


def _from_file(path: str, typ: Any) -> Any:
    with open(path, "r") as f:
        raw_config = yaml.load(f, Loader=yaml.Loader)
        return typ(**raw_config)


class GenericQDeviceConfig(BaseModel):
    # total number of qubits
    num_qubits: int = 2
    # number of communication qubits
    num_comm_qubits: int = 2

    # coherence times (same for each qubit)
    T1: int = 10_000_000_000
    T2: int = 1_000_000_000

    # gate execution times
    init_time: int = 10_000
    single_qubit_gate_time: int = 1_000
    two_qubit_gate_time: int = 100_000
    measure_time: int = 10_000

    # noise model
    single_qubit_gate_depolar_prob: float = 0.0
    two_qubit_gate_depolar_prob: float = 0.01

    @classmethod
    def from_file(cls, path: str) -> GenericQDeviceConfig:
        return _from_file(path, GenericQDeviceConfig)  # type: ignore

    @classmethod
    def perfect_config(cls, num_qubits: int) -> GenericQDeviceConfig:
        cfg = GenericQDeviceConfig(num_qubits=num_qubits, num_comm_qubits=num_qubits)
        cfg.single_qubit_gate_depolar_prob = 0.0
        cfg.two_qubit_gate_depolar_prob = 0.0
        return cfg


class NVQDeviceConfig(BaseModel):
    # number of qubits per NV
    num_qubits: int = 2

    # initialization error of the electron spin
    electron_init_depolar_prob: float = 0.05

    # error of the single-qubit gate
    electron_single_qubit_depolar_prob: float = 0.0

    # measurement errors (prob_error_X is the probability that outcome X is flipped to 1 - X)
    prob_error_0: float = 0.05
    prob_error_1: float = 0.005

    # initialization error of the carbon nuclear spin
    carbon_init_depolar_prob: float = 0.05

    # error of the Z-rotation gate on the carbon nuclear spin
    carbon_z_rot_depolar_prob: float = 0.001

    # error of the native NV two-qubit gate
    ec_gate_depolar_prob: float = 0.008

    # coherence times
    electron_T1: int = 1_000_000_000
    electron_T2: int = 300_000_000
    carbon_T1: int = 150_000_000_000
    carbon_T2: int = 1_500_000_000

    # gate execution times
    carbon_init: int = 310_000
    carbon_rot_x: int = 500_000
    carbon_rot_y: int = 500_000
    carbon_rot_z: int = 500_000
    electron_init: int = 2_000
    electron_rot_x: int = 5_000
    electron_rot_y: int = 5_000
    electron_rot_z: int = 5_000
    ec_controlled_dir_x: int = 500_000
    ec_controlled_dir_y: int = 500_000
    measure: int = 3_700

    @classmethod
    def from_file(cls, path: str) -> NVQDeviceConfig:
        return _from_file(path, NVQDeviceConfig)  # type: ignore

    @classmethod
    def perfect_config(cls, num_qubits: int) -> NVQDeviceConfig:
        # get default config
        cfg = NVQDeviceConfig(num_qubits=num_qubits)

        # set all error params to 0
        cfg.electron_init_depolar_prob = 0
        cfg.electron_single_qubit_depolar_prob = 0
        cfg.prob_error_0 = 0
        cfg.prob_error_1 = 0
        cfg.carbon_init_depolar_prob = 0
        cfg.carbon_z_rot_depolar_prob = 0
        cfg.ec_gate_depolar_prob = 0
        return cfg


class ProcNodeConfig(BaseModel):
    name: str
    node_id: int
    qdevice_typ: str
    qdevice_cfg: Any
    host_qnos_latency: float = 0.0
    instr_latency: float = 0.0
    receive_latency: float = 0.0

    @classmethod
    def from_file(cls, path: str) -> ProcNodeConfig:
        return _from_file(path, ProcNodeConfig)  # type: ignore

    @classmethod
    def perfect_generic_config(
        cls, name: str, node_id: int, num_qubits: int
    ) -> ProcNodeConfig:
        return ProcNodeConfig(
            name=name,
            node_id=node_id,
            qdevice_typ="generic",
            qdevice_cfg=GenericQDeviceConfig.perfect_config(num_qubits),
            host_qnos_latency=0.0,
            instr_latency=0.0,
            receive_latency=0.0,
        )


class DepolariseLinkConfig(BaseModel):
    fidelity: float
    prob_success: float
    t_cycle: float

    @classmethod
    def from_file(cls, path: str) -> DepolariseLinkConfig:
        return _from_file(path, DepolariseLinkConfig)  # type: ignore


class NVLinkConfig(BaseModel):
    length_A: float
    length_B: float
    full_cycle: float
    cycle_time: float
    alpha: float

    @classmethod
    def from_file(cls, path: str) -> NVLinkConfig:
        return _from_file(path, NVLinkConfig)  # type: ignore


class HeraldedLinkConfig(BaseModel):
    length: float
    p_loss_init: float = 0
    p_loss_length: float = 0.25
    speed_of_light: float = 200_000
    dark_count_probability: float = 0
    detector_efficiency: float = 1.0
    visibility: float = 1.0
    num_resolving: bool = False

    @classmethod
    def from_file(cls, path: str) -> HeraldedLinkConfig:
        return _from_file(path, HeraldedLinkConfig)  # type: ignore


class LinkConfig(BaseModel):
    node1: str
    node2: str
    typ: str
    cfg: Any
    host_host_latency: float = 0.0
    qnos_qnos_latency: float = 0.0

    @classmethod
    def from_file(cls, path: str) -> LinkConfig:
        return _from_file(path, LinkConfig)  # type: ignore

    @classmethod
    def perfect_config(cls, node1: str, node2: str) -> LinkConfig:
        return LinkConfig(
            node1=node1,
            node2=node2,
            typ="perfect",
            cfg=None,
            host_host_latency=0.0,
            qnos_qnos_latency=0.0,
        )


class ProcNodeNetworkConfig(BaseModel):
    nodes: List[ProcNodeConfig]
    links: List[LinkConfig]

    @classmethod
    def from_file(cls, path: str) -> ProcNodeNetworkConfig:
        return _from_file(path, ProcNodeNetworkConfig)  # type: ignore
